# -*- coding: utf-8 -*-
'''
Use three emotional dimensions - valence, arousal and dominance - to describe human perceptions of physical environments.

Interpretations of pleasure: Positive versus negative affective states (e.g. excitement, relaxation, love, and
tranquility versus cruelty, humiliation, disinterest, and boredom)

Interpretations of arousal: Level of mental alertness and physical activity. (e.g. sleep, inactivity, boredom, and
relaxation at the lower end versus wakefulness, bodily tension, strenuous
exercise, and concentration at the higher end).

Interpretations of dominance: Ranges from feelings of total lack control or influence on events and surroundings to
the opposite extreme of feeling influential and in control

'''
from __future__ import print_function
import os
import warnings

from engine.object_detection_branch.retina_net.single_img_inference import RetinaNet_single_img_detection
from engine.object_detection_branch.ssd_detector import single_shot_detector

from applications.emotic_utils import _obtain_single_model_VAD,prepare_input_data, _obtain_nb_classifiers, _obtain_ensembling_weights,\
    _obtain_two_models_ensembling_VAD,_obtain_three_models_ensembling_VAD

from scipy.misc import imread
from matplotlib import pyplot as plt
from utils.generic_utils import crop, round_number




def single_img_VAD_inference(img_path,
                             object_detector_backend,
                             model_a_backend_name,
                             model_b_backend_name = None,
                             model_c_backend_name = None):
    """Performs single image inference.
        It also saves the original image (`img_path`) with the overlaid recognised humans bounding boxes and their VAD values.

    # Arguments
        img_path: Path to image file
        object_detector_backend: Backend with which the objects will be detected. One of `SSD` or `RetinaNet`.
            the person who’s feelings are to be estimated.
        model_backend_name: One of `VGG16`, `VGG19` or `ResNet50`.
            Note that EMOTIC model has already combined `model_backend_name` features with `VGG16_Places365` features at training stage,
            but for simplicity reasons only the body backbone CNN name is adjustable.

    # Returns
        Three integer values corresponding to `valence`, `arousal` and `dominance`.

    # Raises
        ImportError: if PIL is not available.
        ValueError: if interpolation method is not supported.
    """

    if not (object_detector_backend in {'SSD', 'RetinaNet'}):
        raise ValueError('The `object_detector_backend_name` argument should be either '
                         '`SSD` for Single-Shot MultiBox Detector or `RetinaNet` for RetinaNet dense detector. ')

    (head, tail) = os.path.split(img_path)
    filename_only = os.path.splitext(tail)[0]

    nb_classifiers, classifiers_names = _obtain_nb_classifiers(model_a_name=model_a_backend_name,
                                                               model_b_name=model_b_backend_name,
                                                               model_c_name=model_c_backend_name)

    save_as = 'results/'+filename_only + '_' + classifiers_names + '.png'

    if nb_classifiers == 1:
        model_a = _obtain_single_model_VAD(model_a_backend_name)

    elif nb_classifiers == 2:
        w_model_a, w_model_b = _obtain_ensembling_weights(nb_classifiers=nb_classifiers,
                                                          model_a_name=model_a_backend_name,
                                                          model_b_name=model_b_backend_name,
                                                          model_c_name=model_c_backend_name)

        model_a, model_b = _obtain_two_models_ensembling_VAD(model_a_name=model_a_backend_name,
                                                             model_b_name=model_b_backend_name)

    elif nb_classifiers == 3:
        w_model_a, w_model_b, w_model_c = _obtain_ensembling_weights(nb_classifiers=nb_classifiers,
                                                                     model_a_name=model_a_backend_name,
                                                                     model_b_name=model_b_backend_name,
                                                                     model_c_name=model_c_backend_name)

        model_a, model_b, model_c = _obtain_three_models_ensembling_VAD(model_a_name=model_a_backend_name,
                                                                        model_b_name=model_b_backend_name,
                                                                        model_c_name=model_c_backend_name)


    numpy_img_path = imread(img_path)

    # ~Object detection branch~
    if object_detector_backend == 'SSD':
        coordinates, persons = single_shot_detector(img_path=img_path, imshow=False)

    elif object_detector_backend == 'RetinaNet':
        coordinates, persons = RetinaNet_single_img_detection(img_path=img_path, imshow=False)

    # configure colours for bounding box and text
    bounding_box_colour_rgbvar = (53, 42, 146)
    bounding_box_colour_rgbvar2 = [x / 255.0 for x in bounding_box_colour_rgbvar]

    text_colour_rgbvar = (214, 86, 100)
    text_colour_rgbvar2 = [x / 255.0 for x in text_colour_rgbvar]

    if persons != 0:
        print('--IMAGE INFERENCE FOR |%d| PERSON(S) FOUND:' % persons)

        plt.figure(figsize=(10, 12))
        plt.imshow(numpy_img_path)

        current_axis = plt.gca()

        counter = 1
        valence_sum = 0
        arousal_sum = 0
        dominance_sum = 0

        for box in coordinates:

            # checks if the number of persons have been reached in order to stop the for loop.
            # if counter > persons:
            #     break

            if box[0] != 0:
                print('[INFO] Person #%d' % counter)

                crop(image_path=img_path, coords=box, saved_location='body_img.jpg')

                x1, x2 = prepare_input_data(body_path = 'body_img.jpg',
                                            image_path = img_path)

                if nb_classifiers == 1:

                    preds = model_a.predict([x1, x2])


                elif nb_classifiers == 2:
                    # obtain predictions
                    preds_model_a = model_a.predict([x1, x2])
                    preds_model_b = model_b.predict([x1, x2])

                    if w_model_a is None and w_model_b is None:
                        # This new prediction array should be more accurate than any of the initial ones
                        preds = 0.50 * (preds_model_a + preds_model_b)

                    else:
                        preds = w_model_a * preds_model_a + w_model_b * preds_model_b

                elif nb_classifiers == 3:
                    # obtain predictions
                    preds_model_a = model_a.predict([x1, x2])
                    preds_model_b = model_b.predict([x1, x2])
                    preds_model_c = model_c.predict([x1, x2])

                    if w_model_a is None and w_model_b is None and w_model_c is None:
                        # This new prediction array should be more accurate than any of the initial ones
                        preds = 0.33 * (preds_model_a + preds_model_b + preds_model_c)

                    else:
                        preds = w_model_a * preds_model_a + w_model_b * preds_model_b + w_model_c * preds_model_c

                # Uncomment to round predicted values
                # valence = round_number(preds[0][0])
                # arousal = round_number(preds[0][1])
                # dominance = round_number(preds[0][2])

                valence = preds[0][0]
                arousal = preds[0][1]
                dominance = preds[0][2]

                print('  Valence (V) --  how pleasant the emotions are: ', valence)
                print('  Arousal (A) --  unrest level of the person(s): ', arousal)
                print('Dominance (D) -- control level of the situation: ', dominance)

                valence_sum += valence
                arousal_sum += arousal
                dominance_sum += dominance

                # current_axis.add_patch(
                #     plt.Rectangle((box[0], box[1]), box[2] - box[0], box[3] - box[1],
                #                   color=text_colour_rgbvar2,
                #                   fill=False,
                #                   linewidth=3.5))




                counter += 1

        global_valence = valence_sum/persons
        global_arousal = arousal_sum/persons
        global_dominance = dominance_sum/persons

        print ('\n')
        print('--GLOBAL EMOTIONAL TRAITS:')

        print("  Valence (V) --  how pleasant the emotions are: %.2f"  % global_valence)
        print("  Arousal (A) --  unrest level of the person(s): %.2f"  % global_arousal)
        print("Dominance (D) --  control level of the situation: %.2f" % global_dominance)
        # print('  Valence (V) --  how pleasant the emotions are: ', global_valence)
        # print('  Arousal (A) --  unrest level of the person(s): ', global_arousal)
        # print('Dominance (D) -- control level of the situation: ', global_dominance)
        #
        # overlayed_text = 'Global emotional traits:' + '\n' '(V): ' + str(round(global_valence,2)) + '\n' '(A): ' + str(round(global_arousal,2)) + '\n' '(D): ' + \
        #                  str(round(global_dominance,2))

        overlayed_text = 'DOMINANCE: ' + \
                         str(round(global_dominance,2))

        current_axis.text(5, -10, overlayed_text, size='x-large', color='white',
                          bbox={'facecolor': bounding_box_colour_rgbvar2, 'alpha': 1.0})




        plt.axis('off')
        plt.savefig(save_as)
        plt.show()
        os.remove("body_img.jpg")

    else:
        warnings.warn('No global emotional traits were identified: '
                      'there was no person detected in the image.')

        global_valence = 0
        global_arousal = 0
        global_dominance = 0



    return global_valence, global_arousal, global_dominance





def single_img_VAD_inference_return_only(img_path,
                                         object_detector_backend,
                                         model_a_backend_name,
                                         model_b_backend_name=None,
                                         model_c_backend_name=None):
    """Performs single image inference.

    # Arguments
        img_path: Path to image file
        object_detector_backend: Backend with which the objects will be detected. One of `SSD` or `RetinaNet`.
            the person who’s feelings are to be estimated.
        model_backend_name: One of `VGG16`, `VGG19` or `ResNet50`.
            Note that EMOTIC model has already combined `model_backend_name` features with `VGG16_Places365` features at training stage,
            but for simplicity reasons only the body backbone CNN name is adjustable.

    # Returns
        Three integer values corresponding to `valence`, `arousal` and `dominance`.

    # Raises
        ImportError: if PIL is not available.
        ValueError: if interpolation method is not supported.
    """

    if not (object_detector_backend in {'SSD', 'RetinaNet'}):
        raise ValueError('The `object_detector_backend_name` argument should be either '
                         '`SSD` for Single-Shot MultiBox Detector or `RetinaNet` for RetinaNet dense detector. ')


    nb_classifiers, classifiers_names = _obtain_nb_classifiers(model_a_name=model_a_backend_name,
                                                               model_b_name=model_b_backend_name,
                                                               model_c_name=model_c_backend_name)


    if nb_classifiers == 1:
        model_a = _obtain_single_model_VAD(model_a_backend_name)

    elif nb_classifiers == 2:
        w_model_a, w_model_b = _obtain_ensembling_weights(nb_classifiers=nb_classifiers,
                                                          model_a_name=model_a_backend_name,
                                                          model_b_name=model_b_backend_name,
                                                          model_c_name=model_c_backend_name)

        model_a, model_b = _obtain_two_models_ensembling_VAD(model_a_name=model_a_backend_name,
                                                             model_b_name=model_b_backend_name)

    elif nb_classifiers == 3:
        w_model_a, w_model_b, w_model_c = _obtain_ensembling_weights(nb_classifiers=nb_classifiers,
                                                                     model_a_name=model_a_backend_name,
                                                                     model_b_name=model_b_backend_name,
                                                                     model_c_name=model_c_backend_name)

        model_a, model_b, model_c = _obtain_three_models_ensembling_VAD(model_a_name=model_a_backend_name,
                                                                        model_b_name=model_b_backend_name,
                                                                        model_c_name=model_c_backend_name)
    # Uncomment for extra verbosity
    # print('[INFO] EMOTIC VAD models have been loaded')

    # numpy_img_path = imread(img_path)

    # ~Object detection branch~
    if object_detector_backend == 'SSD':
        coordinates, persons = single_shot_detector(img_path=img_path, imshow=False)

    elif object_detector_backend == 'RetinaNet':
        coordinates, persons = RetinaNet_single_img_detection(img_path=img_path, imshow=False)

    # Uncomment for extra verbosity
    # print('[INFO] Objects in image have been detected')


    if persons != 0:
        # Uncomment for extra verbosity
        # print('[INFO] Carrying out continuous emotion recognition in VAD space for %d person(s) found: ' % persons)

        counter = 1
        dominance_sum = 0
        valence_sum = 0

        for box in coordinates:

            # checks if the number of persons have been reached in order to stop the for loop.
            # if counter > persons:
            #     break

            if box[0] != 0:
                # Uncomment for extra verbosity
                # print('[INFO] Person #%d' % counter)

                crop(image_path=img_path, coords=box, saved_location='body_img.jpg')

                x1, x2 = prepare_input_data(body_path = 'body_img.jpg',
                                            image_path = img_path)

                if nb_classifiers == 1:

                    preds = model_a.predict([x1, x2])


                elif nb_classifiers == 2:
                    # obtain predictions
                    preds_model_a = model_a.predict([x1, x2])
                    preds_model_b = model_b.predict([x1, x2])

                    if w_model_a is None and w_model_b is None:
                        # This new prediction array should be more accurate than any of the initial ones
                        preds = 0.50 * (preds_model_a + preds_model_b)

                    else:
                        preds = w_model_a * preds_model_a + w_model_b * preds_model_b

                elif nb_classifiers == 3:
                    # obtain predictions
                    preds_model_a = model_a.predict([x1, x2])
                    preds_model_b = model_b.predict([x1, x2])
                    preds_model_c = model_c.predict([x1, x2])

                    if w_model_a is None and w_model_b is None and w_model_c is None:
                        # This new prediction array should be more accurate than any of the initial ones
                        preds = 0.33 * (preds_model_a + preds_model_b + preds_model_c)

                    else:
                        preds = w_model_a * preds_model_a + w_model_b * preds_model_b + w_model_c * preds_model_c

                # Uncomment to round predicted values
                # valence = round_number(preds[0][0])
                # arousal = round_number(preds[0][1])
                # dominance = round_number(preds[0][2])

                valence = preds[0][0]
                # arousal = preds[0][1]
                dominance = preds[0][2]

                # Uncomment for extra verbosity
                # print('  Valence (V): ', valence)
                # print('  Arousal (A): ', arousal)
                # print('Dominance (D): ', dominance)

                valence_sum += valence
                # arousal_sum += arousal
                dominance_sum += dominance



                counter += 1

        global_valence = valence_sum/persons
        # global_arousal = arousal_sum/persons
        global_dominance = dominance_sum/persons

        # Uncomment for extra verbosity
        # print ('\n')
        # print('[INFO] Global emotional traits::')
        # print('  Valence (V) --  how pleasant the emotions are: ', global_valence)
        # print('  Arousal (A) --  unrest level of the person(s): ', global_arousal)
        # print('Dominance (D) -- control level of the situation: ', global_dominance)
        # print('\n')

        os.remove("body_img.jpg")

    else:
        print("[WARNING] No global emotional traits were identified -- no `people` found in input image `", img_path, '`')

        global_valence = 0
        # global_arousal = 0
        global_dominance = 0



    return global_valence, global_dominance


def single_img_VAD_inference_with_bounding_boxes(img_path,
                                                 object_detector_backend,
                                                 model_a_backend_name,
                                                 model_b_backend_name=None,
                                                 model_c_backend_name=None):
    """Performs single image inference.
        It also saves the original image (`img_path`) with the overlaid recognised humans bounding boxes and their VAD values.

    # Arguments
        img_path: Path to image file
        object_detector_backend: Backend with which the objects will be detected. One of `SSD` or `RetinaNet`.
            the person who’s feelings are to be estimated.
        model_backend_name: One of `VGG16`, `VGG19` or `ResNet50`.
            Note that EMOTIC model has already combined `model_backend_name` features with `VGG16_Places365` features at training stage,
            but for simplicity reasons only the body backbone CNN name is adjustable.

    # Returns
        Three integer values corresponding to `valence`, `arousal` and `dominance`.

    # Raises
        ImportError: if PIL is not available.
        ValueError: if interpolation method is not supported.
    """

    if not (object_detector_backend in {'SSD', 'RetinaNet'}):
        raise ValueError('The `object_detector_backend_name` argument should be either '
                         '`SSD` for Single-Shot MultiBox Detector or `RetinaNet` for RetinaNet dense detector. ')

    (head, tail) = os.path.split(img_path)
    filename_only = os.path.splitext(tail)[0]

    nb_classifiers, classifiers_names = _obtain_nb_classifiers(model_a_name=model_a_backend_name,
                                                               model_b_name=model_b_backend_name,
                                                               model_c_name=model_c_backend_name)

    save_as = 'results/'+filename_only + '_' + classifiers_names + '.png'

    if nb_classifiers == 1:
        model_a = _obtain_single_model_VAD(model_a_backend_name)

    elif nb_classifiers == 2:
        w_model_a, w_model_b = _obtain_ensembling_weights(nb_classifiers=nb_classifiers,
                                                          model_a_name=model_a_backend_name,
                                                          model_b_name=model_b_backend_name,
                                                          model_c_name=model_c_backend_name)

        model_a, model_b = _obtain_two_models_ensembling_VAD(model_a_name=model_a_backend_name,
                                                             model_b_name=model_b_backend_name)

    elif nb_classifiers == 3:
        w_model_a, w_model_b, w_model_c = _obtain_ensembling_weights(nb_classifiers=nb_classifiers,
                                                                     model_a_name=model_a_backend_name,
                                                                     model_b_name=model_b_backend_name,
                                                                     model_c_name=model_c_backend_name)

        model_a, model_b, model_c = _obtain_three_models_ensembling_VAD(model_a_name=model_a_backend_name,
                                                                        model_b_name=model_b_backend_name,
                                                                        model_c_name=model_c_backend_name)


    numpy_img_path = imread(img_path)

    # ~Object detection branch~
    if object_detector_backend == 'SSD':
        coordinates, persons = single_shot_detector(img_path=img_path, imshow=False)

    elif object_detector_backend == 'RetinaNet':
        coordinates, persons = RetinaNet_single_img_detection(img_path=img_path, imshow=False)

    # configure colours for bounding box and text
    bounding_box_colour_rgbvar = (53, 42, 146)
    bounding_box_colour_rgbvar2 = [x / 255.0 for x in bounding_box_colour_rgbvar]

    text_colour_rgbvar = (214, 86, 100)
    text_colour_rgbvar2 = [x / 255.0 for x in text_colour_rgbvar]

    if persons != 0:
        print('--IMAGE INFERENCE FOR |%d| PERSON(S) FOUND:' % persons)

        plt.figure(figsize=(10, 12))
        plt.imshow(numpy_img_path)

        current_axis = plt.gca()

        counter = 1
        valence_sum = 0
        arousal_sum = 0
        dominance_sum = 0

        for box in coordinates:

            # checks if the number of persons have been reached in order to stop the for loop.
            # if counter > persons:
            #     break

            if box[0] != 0:
                print('[INFO] Person #%d' % counter)

                crop(image_path=img_path, coords=box, saved_location='body_img.jpg')

                x1, x2 = prepare_input_data(body_path = 'body_img.jpg',
                                            image_path = img_path)

                if nb_classifiers == 1:

                    preds = model_a.predict([x1, x2])


                elif nb_classifiers == 2:
                    # obtain predictions
                    preds_model_a = model_a.predict([x1, x2])
                    preds_model_b = model_b.predict([x1, x2])

                    if w_model_a is None and w_model_b is None:
                        # This new prediction array should be more accurate than any of the initial ones
                        preds = 0.50 * (preds_model_a + preds_model_b)

                    else:
                        preds = w_model_a * preds_model_a + w_model_b * preds_model_b

                elif nb_classifiers == 3:
                    # obtain predictions
                    preds_model_a = model_a.predict([x1, x2])
                    preds_model_b = model_b.predict([x1, x2])
                    preds_model_c = model_c.predict([x1, x2])

                    if w_model_a is None and w_model_b is None and w_model_c is None:
                        # This new prediction array should be more accurate than any of the initial ones
                        preds = 0.33 * (preds_model_a + preds_model_b + preds_model_c)

                    else:
                        preds = w_model_a * preds_model_a + w_model_b * preds_model_b + w_model_c * preds_model_c

                # Uncomment to round predicted values
                # valence = round_number(preds[0][0])
                # arousal = round_number(preds[0][1])
                # dominance = round_number(preds[0][2])

                valence = preds[0][0]
                arousal = preds[0][1]
                dominance = preds[0][2]

                print('  Valence (V) --  how pleasant the emotions are: ', valence)
                print('  Arousal (A) --  unrest level of the person(s): ', arousal)
                print('Dominance (D) -- control level of the situation: ', dominance)

                valence_sum += valence
                arousal_sum += arousal
                dominance_sum += dominance

                current_axis.add_patch(
                    plt.Rectangle((box[0], box[1]), box[2] - box[0], box[3] - box[1],
                                  color=text_colour_rgbvar2,
                                  fill=False,
                                  linewidth=3.5))

                people_VAD_overlayed_text = '(V): ' + str(round(valence, 2)) + '\n' '(A): ' \
                                            + str(round(arousal, 2)) + '\n' '(D): ' \
                                            + str(round(dominance, 2))

                current_axis.text(box[0]+5, box[1]-10, people_VAD_overlayed_text, size='x-large', color='white',
                                  bbox={'facecolor': bounding_box_colour_rgbvar2, 'alpha': 1.0})




                counter += 1

        global_valence = valence_sum/persons
        global_arousal = arousal_sum/persons
        global_dominance = dominance_sum/persons

        print ('\n')
        print('--GLOBAL EMOTIONAL TRAITS:')

        print("  Valence (V) --  how pleasant the emotions are: %.2f"  % global_valence)
        print("  Arousal (A) --  unrest level of the person(s): %.2f"  % global_arousal)
        print("Dominance (D) --  control level of the situation: %.2f" % global_dominance)
        # print('  Valence (V) --  how pleasant the emotions are: ', global_valence)
        # print('  Arousal (A) --  unrest level of the person(s): ', global_arousal)
        # print('Dominance (D) -- control level of the situation: ', global_dominance)

        overlayed_text = '(V): ' + str(round(global_valence,2)) + '\n' '(A): ' + str(round(global_arousal,2)) + '\n' '(D): ' + \
                         str(round(global_dominance,2))


        # current_axis.text(0, 0, overlayed_text, size='x-large', color='white',
        #                   bbox={'facecolor': bounding_box_colour_rgbvar2, 'alpha': 1.0})

        plt.axis('off')
        plt.savefig(save_as)
        plt.show()
        os.remove("body_img.jpg")

    else:
        warnings.warn('No global emotional traits were identified: '
                      'there was no person detected in the image.')

        global_valence = 0
        global_arousal = 0
        global_dominance = 0



    return global_valence, global_arousal, global_dominance





if __name__ == "__main__":

    img_path = '/home/sandbox/Desktop/Two-class-HRV/ChildLabour/test/no_child_labour/no_child_labour_0015.jpg'
    model_a_backend_name = 'VGG19'
    model_b_backend_name = 'VGG16'
    model_c_backend_name = 'ResNet50'

    valence, arousal, dominance = single_img_VAD_inference(img_path = img_path,
                                                           object_detector_backend='RetinaNet',
                                                           model_a_backend_name = model_a_backend_name,
                                                           model_b_backend_name=model_b_backend_name,
                                                           model_c_backend_name=model_c_backend_name,
                                                           )


"""
Copyright 2017-2018 Fizyr (https://fizyr.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np

import keras

from .. import initializers
from .. import layers


def default_classification_model(
    num_classes,
    num_anchors,
    pyramid_feature_size=256,
    prior_probability=0.01,
    classification_feature_size=256,
    name='classification_submodel'
):
    """ Creates the default regression submodel.

    Args
        num_classes                 : Number of classes to predict a score for at each feature level.
        num_anchors                 : Number of anchors to predict classification scores for at each feature level.
        pyramid_feature_size        : The number of filters to expect from the feature pyramid levels.
        classification_feature_size : The number of filters to use in the layers in the classification submodel.
        name                        : The name of the submodel.

    Returns
        A keras.models.Model that predicts classes for each anchor.
    """
    options = {
        'kernel_size' : 3,
        'strides'     : 1,
        'padding'     : 'same',
    }

    inputs  = keras.layers.Input(shape=(None, None, pyramid_feature_size))
    outputs = inputs
    for i in range(4):
        outputs = keras.layers.Conv2D(
            filters=classification_feature_size,
            activation='relu',
            name='pyramid_classification_{}'.format(i),
            kernel_initializer=keras.initializers.normal(mean=0.0, stddev=0.01, seed=None),
            bias_initializer='zeros',
            **options
        )(outputs)

    outputs = keras.layers.Conv2D(
        filters=num_classes * num_anchors,
        kernel_initializer=keras.initializers.zeros(),
        bias_initializer=initializers.PriorProbability(probability=prior_probability),
        name='pyramid_classification',
        **options
    )(outputs)

    # reshape output and apply sigmoid
    outputs = keras.layers.Reshape((-1, num_classes), name='pyramid_classification_reshape')(outputs)
    outputs = keras.layers.Activation('sigmoid', name='pyramid_classification_sigmoid')(outputs)

    return keras.models.Model(inputs=inputs, outputs=outputs, name=name)


def default_regression_model(num_anchors, pyramid_feature_size=256, regression_feature_size=256, name='regression_submodel'):
    """ Creates the default regression submodel.

    Args
        num_anchors             : Number of anchors to regress for each feature level.
        pyramid_feature_size    : The number of filters to expect from the feature pyramid levels.
        regression_feature_size : The number of filters to use in the layers in the regression submodel.
        name                    : The name of the submodel.

    Returns
        A keras.models.Model that predicts regression values for each anchor.
    """
    # All new conv layers except the final one in the
    # RetinaNet (classification) subnets are initialized
    # with bias b = 0 and a Gaussian weight fill with stddev = 0.01.
    options = {
        'kernel_size'        : 3,
        'strides'            : 1,
        'padding'            : 'same',
        'kernel_initializer' : keras.initializers.normal(mean=0.0, stddev=0.01, seed=None),
        'bias_initializer'   : 'zeros'
    }

    inputs  = keras.layers.Input(shape=(None, None, pyramid_feature_size))
    outputs = inputs
    for i in range(4):
        outputs = keras.layers.Conv2D(
            filters=regression_feature_size,
            activation='relu',
            name='pyramid_regression_{}'.format(i),
            **options
        )(outputs)

    outputs = keras.layers.Conv2D(num_anchors * 4, name='pyramid_regression', **options)(outputs)
    outputs = keras.layers.Reshape((-1, 4), name='pyramid_regression_reshape')(outputs)

    return keras.models.Model(inputs=inputs, outputs=outputs, name=name)


def __create_pyramid_features(C3, C4, C5, feature_size=256):
    """ Creates the FPN layers on top of the backbone features.

    Args
        C3           : Feature stage C3 from the backbone.
        C4           : Feature stage C4 from the backbone.
        C5           : Feature stage C5 from the backbone.
        feature_size : The feature size to use for the resulting feature levels.

    Returns
        A list of feature levels [P3, P4, P5, P6, P7].
    """
    # upsample C5 to get P5 from the FPN paper
    P5           = keras.layers.Conv2D(feature_size, kernel_size=1, strides=1, padding='same', name='C5_reduced')(C5)
    P5_upsampled = layers.UpsampleLike(name='P5_upsampled')([P5, C4])
    P5           = keras.layers.Conv2D(feature_size, kernel_size=3, strides=1, padding='same', name='P5')(P5)

    # add P5 elementwise to C4
    P4           = keras.layers.Conv2D(feature_size, kernel_size=1, strides=1, padding='same', name='C4_reduced')(C4)
    P4           = keras.layers.Add(name='P4_merged')([P5_upsampled, P4])
    P4_upsampled = layers.UpsampleLike(name='P4_upsampled')([P4, C3])
    P4           = keras.layers.Conv2D(feature_size, kernel_size=3, strides=1, padding='same', name='P4')(P4)

    # add P4 elementwise to C3
    P3 = keras.layers.Conv2D(feature_size, kernel_size=1, strides=1, padding='same', name='C3_reduced')(C3)
    P3 = keras.layers.Add(name='P3_merged')([P4_upsampled, P3])
    P3 = keras.layers.Conv2D(feature_size, kernel_size=3, strides=1, padding='same', name='P3')(P3)

    # "P6 is obtained via a 3x3 stride-2 conv on C5"
    P6 = keras.layers.Conv2D(feature_size, kernel_size=3, strides=2, padding='same', name='P6')(C5)

    # "P7 is computed by applying ReLU followed by a 3x3 stride-2 conv on P6"
    P7 = keras.layers.Activation('relu', name='C6_relu')(P6)
    P7 = keras.layers.Conv2D(feature_size, kernel_size=3, strides=2, padding='same', name='P7')(P7)

    return [P3, P4, P5, P6, P7]


class AnchorParameters:
    """ The parameteres that define how anchors are generated.

    Args
        sizes   : List of sizes to use. Each size corresponds to one feature level.
        strides : List of strides to use. Each stride correspond to one feature level.
        ratios  : List of ratios to use per location in a feature map.
        scales  : List of scales to use per location in a feature map.
    """
    def __init__(self, sizes, strides, ratios, scales):
        self.sizes   = sizes
        self.strides = strides
        self.ratios  = ratios
        self.scales  = scales

    def num_anchors(self):
        return len(self.ratios) * len(self.scales)


"""
The default anchor parameters.
"""
AnchorParameters.default = AnchorParameters(
    sizes   = [32, 64, 128, 256, 512],
    strides = [8, 16, 32, 64, 128],
    ratios  = np.array([0.5, 1, 2], keras.backend.floatx()),
    scales  = np.array([2 ** 0, 2 ** (1.0 / 3.0), 2 ** (2.0 / 3.0)], keras.backend.floatx()),
)


def default_submodels(num_classes, num_anchors):
    """ Create a list of default submodels used for object detection.

    The default submodels contains a regression submodel and a classification submodel.

    Args
        num_classes : Number of classes to use.
        num_anchors : Number of base anchors.

    Returns
        A list of tuple, where the first element is the name of the submodel and the second element is the submodel itself.
    """
    return [
        ('regression', default_regression_model(num_anchors)),
        ('classification', default_classification_model(num_classes, num_anchors))
    ]


def __build_model_pyramid(name, model, features):
    """ Applies a single submodel to each FPN level.

    Args
        name     : Name of the submodel.
        model    : The submodel to evaluate.
        features : The FPN features.

    Returns
        A tensor containing the response from the submodel on the FPN features.
    """
    return keras.layers.Concatenate(axis=1, name=name)([model(f) for f in features])


def __build_pyramid(models, features):
    """ Applies all submodels to each FPN level.

    Args
        models   : List of sumodels to run on each pyramid level (by default only regression, classifcation).
        features : The FPN features.

    Returns
        A list of tensors, one for each submodel.
    """
    return [__build_model_pyramid(n, m, features) for n, m in models]


def __build_anchors(anchor_parameters, features):
    """ Builds anchors for the shape of the features from FPN.

    Args
        anchor_parameters : Parameteres that determine how anchors are generated.
        features          : The FPN features.

    Returns
        A tensor containing the anchors for the FPN features.

        The shape is:
        ```
        (batch_size, num_anchors, 4)
        ```
    """
    anchors = [
        layers.Anchors(
            size=anchor_parameters.sizes[i],
            stride=anchor_parameters.strides[i],
            ratios=anchor_parameters.ratios,
            scales=anchor_parameters.scales,
            name='anchors_{}'.format(i)
        )(f) for i, f in enumerate(features)
    ]

    return keras.layers.Concatenate(axis=1, name='anchors')(anchors)


def retinanet(
    inputs,
    backbone_layers,
    num_classes,
    num_anchors             = 9,
    create_pyramid_features = __create_pyramid_features,
    submodels               = None,
    name                    = 'retinanet'
):
    """ Construct a RetinaNet model on top of a backbone.

    This model is the minimum model necessary for training (with the unfortunate exception of anchors as output).

    Args
        inputs                  : keras.layers.Input (or list of) for the input to the model.
        num_classes             : Number of classes to classify.
        num_anchors             : Number of base anchors.
        create_pyramid_features : Functor for creating pyramid features given the features C3, C4, C5 from the backbone.
        submodels               : Submodels to run on each feature map (default is regression and classification submodels).
        name                    : Name of the model.

    Returns
        A keras.models.Model which takes an image as input and outputs generated anchors and the result from each submodel on every pyramid level.

        The order of the outputs is as defined in submodels:
        ```
        [
            regression, classification, other[0], other[1], ...
        ]
        ```
    """
    if submodels is None:
        submodels = default_submodels(num_classes, num_anchors)

    C3, C4, C5 = backbone_layers

    # compute pyramid features as per https://arxiv.org/abs/1708.02002
    features = create_pyramid_features(C3, C4, C5)

    # for all pyramid levels, run available submodels
    pyramids = __build_pyramid(submodels, features)

    return keras.models.Model(inputs=inputs, outputs=pyramids, name=name)


def retinanet_bbox(
    model             = None,
    anchor_parameters = AnchorParameters.default,
    nms               = True,
    name              = 'retinanet-bbox',
    **kwargs
):
    """ Construct a RetinaNet model on top of a backbone and adds convenience functions to output boxes directly.

    This model uses the minimum retinanet model and appends a few layers to compute boxes within the graph.
    These layers include applying the regression values to the anchors and performing NMS.

    Args
        model             : RetinaNet model to append bbox layers to. If None, it will create a RetinaNet model using **kwargs.
        anchor_parameters : Struct containing configuration for anchor generation (sizes, strides, ratios, scales).
        name              : Name of the model.
        *kwargs           : Additional kwargs to pass to the minimal retinanet model.

    Returns
        A keras.models.Model which takes an image as input and outputs the detections on the image.

        The order is defined as follows:
        ```
        [
            boxes, scores, labels, other[0], other[1], ...
        ]
        ```
    """
    if model is None:
        model = retinanet(num_anchors=anchor_parameters.num_anchors(), **kwargs)

    # compute the anchors
    features = [model.get_layer(p_name).output for p_name in ['P3', 'P4', 'P5', 'P6', 'P7']]
    anchors  = __build_anchors(anchor_parameters, features)

    # we expect the anchors, regression and classification values as first output
    regression     = model.outputs[0]
    classification = model.outputs[1]

    # "other" can be any additional output from custom submodels, by default this will be []
    other = model.outputs[2:]

    # apply predicted regression to anchors
    boxes = layers.RegressBoxes(name='boxes')([anchors, regression])
    boxes = layers.ClipBoxes(name='clipped_boxes')([model.inputs[0], boxes])

    # filter detections (apply NMS / score threshold / select top-k)
    detections = layers.FilterDetections(nms=nms, name='filtered_detections')([boxes, classification] + other)

    outputs = detections

    # construct the model
    return keras.models.Model(inputs=model.inputs, outputs=outputs, name=name)



"""Generic utilities continuous-emotion-recognition-in-VAD-space-based models.
"""

from __future__ import print_function
import numpy as np

import keras.backend as K
import itertools
from itertools import product
import tensorflow as tf
import keras.backend.tensorflow_backend as tfb
from keras.preprocessing import image
from applications.emotic_vgg16 import EMOTIC_VAD_VGG16
from applications.emotic_vgg19 import EMOTIC_VAD_VGG19
from applications.emotic_resnet50 import EMOTIC_VAD_ResNet50
from utils.generic_utils import imagenet_preprocess_input, places_preprocess_input



target_size = (224, 224)

CLASS_INDEX = None
CLASS_INDEX_PATH = 'https://github.com/GKalliatakis/Keras-EMOTIC-resources/releases/download/v1.0/emotic_class_index.json'


def _obtain_weights_CSVLogger_filenames(body_backbone_CNN, image_backbone_CNN):
    """Obtains the polished filenames for the weights and the CSVLogger of the model.

    # Arguments
        model_name: String to declare the name of the model

    # Returns
        Two strings that will serve as the filenames for the weights and the CSVLogger respectively.
    """

    prefix= 'trained_models/emotic_vad_'
    suffix= '_weights_tf_dim_ordering_tf_kernels.h5'
    weights_filename = prefix + body_backbone_CNN + suffix


    CSVLogger_filename = 'emotic_vad_'+body_backbone_CNN+'_training.csv'

    return weights_filename, CSVLogger_filename



# ----------------------------------------------------------------------------------------------------- #
#                                    Obtain number of classifiers
# ----------------------------------------------------------------------------------------------------- #

def _obtain_nb_classifiers(model_a_name = None, model_b_name = None, model_c_name = None):
    """Obtains the number of different classifiers based on given model names.
        Note that EMOTIC model has already combined body backbone CNN features (which in this case are the  `model_b_name` or `model_c_name`
        features, with `VGG16_Places365` features at training stage, but for simplicity reasons only the body backbone CNN name is adjustable.

    # Arguments
        model_b_name: One of `VGG16`, `VGG19`, `ResNet50` or `None`.
        model_c_name: One of `VGG16`, `VGG19`, `ResNet50` or `None`.

    # Returns
        The number of different classifiers alongside a polished file_name
    """

    if (model_b_name) is None and (model_c_name is not None):
        raise ValueError('The models names must be set in the correct order starting from model_a --> model_b --> model_c. ')


    if (model_b_name in {None}) and (model_c_name in {None}):
        nb_classifiers = 1
        file_name = model_a_name

        return nb_classifiers, file_name

    if (model_b_name in {'VGG16', 'VGG19', 'ResNet50'}) and (model_c_name in {None}):
        nb_classifiers = 2
        file_name = model_a_name + '_' + model_b_name
        return nb_classifiers, file_name

    if (model_b_name in {'VGG16', 'VGG19', 'ResNet50'})  and (model_c_name in {'VGG16', 'VGG19', 'ResNet50'}) :
        nb_classifiers = 3
        file_name = model_a_name + '_' + model_b_name + '_' + model_c_name
        return nb_classifiers, file_name


# ----------------------------------------------------------------------------------------------------- #
#                                    Prepare input images
# ----------------------------------------------------------------------------------------------------- #

def prepare_input_data(body_path,
                       image_path):
    """Prepares the raw images for the EMOTIC model.

    # Arguments
        body_path: Path to body only image file.
        image_path: Path to entire image file.

    # Returns
        The two processed images
    """

    body_img = image.load_img(body_path, target_size=(224, 224))
    x1 = image.img_to_array(body_img)
    x1 = np.expand_dims(x1, axis=0)
    x1 = imagenet_preprocess_input(x1)

    entire_img = image.load_img(image_path, target_size=(224, 224))
    x2 = image.img_to_array(entire_img)
    x2 = np.expand_dims(x2, axis=0)
    x2 = places_preprocess_input(x2)


    return x1, x2


# ----------------------------------------------------------------------------------------------------- #
#                            Obtain ensembling weights for different classifiers
# ----------------------------------------------------------------------------------------------------- #

def _obtain_ensembling_weights(nb_classifiers,
                               model_a_name,
                               model_b_name,
                               model_c_name):
    """Obtains the set of ensembling weights that will be used for conducting weighted average.

    # Arguments
        nb_classifiers: Integer, number of different classifiers that will be used for ensembling weights.
        model_a_name: One of `VGG16`, `VGG19`, `ResNet50` or `None`.
        model_b_name: One of `VGG16`, `VGG19`, `ResNet50` or `None`.
        model_c_name: One of `VGG16`, `VGG19`, `ResNet50` or `None`.

    # Returns
        The weights (float) for every model and every dimension.
    """

    if nb_classifiers == 2:

        if model_a_name == 'VGG16' and model_b_name == 'ResNet50':
            w_model_a = 0.55
            w_model_b = 0.45

            return w_model_a, w_model_b

        # ensure that giving models with different order will not effect the weights
        elif model_a_name == 'ResNet50' and model_b_name == 'VGG16':
            w_model_a = 0.45
            w_model_b = 0.55

            return w_model_a, w_model_b


        elif model_a_name == 'VGG16' and model_b_name == 'VGG19':
            w_model_a = 0.48
            w_model_b = 0.52

            return w_model_a, w_model_b

        # ensure that giving models with different order will not effect the weights
        elif model_a_name == 'VGG19' and model_b_name == 'VGG16':
            w_model_a = 0.52
            w_model_b = 0.48

            return w_model_a, w_model_b


        elif model_a_name == 'ResNet50' and model_b_name == 'VGG19':
            w_model_a = 0.40
            w_model_b = 0.60

            return w_model_a, w_model_b


        # ensure that giving models with different order will not effect the weights
        elif model_a_name == 'VGG19' and model_b_name == 'ResNet50':

            w_model_a = 0.60
            w_model_b = 0.40

            return w_model_a, w_model_b


    elif nb_classifiers == 3:

        if model_a_name == 'VGG16' and model_b_name == 'ResNet50' and model_c_name == 'VGG19':
            w_model_a = 0.35
            w_model_b = 0.28
            w_model_c = 0.37

            return w_model_a, w_model_b, w_model_c

        elif model_a_name == 'VGG16' and model_b_name == 'VGG19' and model_c_name == 'ResNet50':
            w_model_a = 0.35
            w_model_b = 0.37
            w_model_c = 0.28

            return w_model_a, w_model_b, w_model_c


        elif model_a_name == 'ResNet50' and model_b_name == 'VGG16' and model_c_name == 'VGG19':
            w_model_a = 0.28
            w_model_b = 0.35
            w_model_c = 0.37

            return w_model_a, w_model_b, w_model_c


        elif model_a_name == 'ResNet50' and model_b_name == 'VGG19' and model_c_name == 'VGG16':
            w_model_a = 0.28
            w_model_b = 0.37
            w_model_c = 0.35

            return w_model_a, w_model_b, w_model_c

        elif model_a_name == 'VGG19' and model_b_name == 'ResNet50' and model_c_name == 'VGG16':
            w_model_a = 0.37
            w_model_b = 0.28
            w_model_c = 0.35

            return w_model_a, w_model_b, w_model_c


        elif model_a_name == 'VGG19' and model_b_name == 'VGG16' and model_c_name == 'ResNet50':
            w_model_a = 0.37
            w_model_b = 0.35
            w_model_c = 0.28

            return w_model_a, w_model_b, w_model_c



# ----------------------------------------------------------------------------------------------------- #
#                      Obtain Keras model instances based on given model names
# ----------------------------------------------------------------------------------------------------- #

def _obtain_single_model_VAD(model_a_name):
    """Instantiates and returns 1 Keras model instance based on the given model name.

    # Arguments
        model_a_name: String to declare the name of the 1st model

    # Returns
        Single Keras model instance.
    """

    if model_a_name == 'VGG16':
        model_a = EMOTIC_VAD_VGG16(include_top=True, weights='emotic')

    elif model_a_name == 'VGG19':
        model_a = EMOTIC_VAD_VGG19(include_top=True, weights='emotic')

    elif model_a_name == 'ResNet50':
        model_a = EMOTIC_VAD_ResNet50(include_top=True, weights='emotic')

    return model_a


def _obtain_two_models_ensembling_VAD(model_a_name, model_b_name):
    """Instantiates and returns 2 Keras model instances based on the given model names.

    # Arguments
        model_a_name: String to declare the name of the 1st model
        model_b_name: String to declare the name of the 2nd model

    # Returns
        Two Keras model instances.
    """

    if model_a_name == 'VGG16':
        model_a = EMOTIC_VAD_VGG16(include_top=True, weights='emotic')

    elif model_a_name == 'VGG19':
        model_a = EMOTIC_VAD_VGG19(include_top=True, weights='emotic')

    elif model_a_name == 'ResNet50':
        model_a = EMOTIC_VAD_ResNet50(include_top=True, weights='emotic')


    if model_b_name == 'VGG16':
        model_b = EMOTIC_VAD_VGG16(include_top=True, weights='emotic')

    elif model_b_name == 'VGG19':
        model_b = EMOTIC_VAD_VGG19(include_top=True, weights='emotic')

    elif model_b_name == 'ResNet50':
        model_b = EMOTIC_VAD_ResNet50(include_top=True, weights='emotic')

    return model_a, model_b


def _obtain_three_models_ensembling_VAD(model_a_name, model_b_name, model_c_name):
    """Instantiates and returns 3 Keras model instances based on the given model names.

    # Arguments
        model_a_name: String to declare the name of the 1st model
        model_b_name: String to declare the name of the 2nd model
        model_c_name: String to declare the name of the 3rd model

    # Returns
        Three Keras model instances.
    """

    if model_a_name == 'VGG16':
        model_a = EMOTIC_VAD_VGG16(include_top=True, weights='emotic')

    elif model_a_name == 'VGG19':
        model_a = EMOTIC_VAD_VGG19(include_top=True, weights='emotic')

    elif model_a_name == 'ResNet50':
        model_a = EMOTIC_VAD_ResNet50(include_top=True, weights='emotic')


    if model_b_name == 'VGG16':
        model_b = EMOTIC_VAD_VGG16(include_top=True, weights='emotic')

    elif model_b_name == 'VGG19':
        model_b = EMOTIC_VAD_VGG19(include_top=True, weights='emotic')

    elif model_b_name == 'ResNet50':
        model_b = EMOTIC_VAD_ResNet50(include_top=True, weights='emotic')


    if model_c_name == 'VGG16':
        model_c = EMOTIC_VAD_VGG16(include_top=True, weights='emotic')

    elif model_c_name == 'VGG19':
        model_c = EMOTIC_VAD_VGG19(include_top=True, weights='emotic')

    elif model_c_name == 'ResNet50':
        model_c = EMOTIC_VAD_ResNet50(include_top=True, weights='emotic')

    return model_a, model_b, model_c






# -*- coding: utf-8 -*-
'''

'''
from __future__ import print_function
import os
from inference.displacenet_single_image_inference_unified import displaceNet_inference


img_path = '/home/sandbox/GKalliatakis-GitHub-Account/DisplaceNet/displaced_people_test_image.jpg'
violation_class = 'dp'
hra_model_backend_name = 'VGG16'
nb_of_conv_layers_to_fine_tune = 1

emotic_model_a_backend_name = 'VGG16'
emotic_model_b_backend_name = None
emotic_model_c_backend_name = None


DisplaceNet_preds = displaceNet_inference(img_path=img_path,
                                          emotic_model_a_backend_name=emotic_model_a_backend_name,
                                          emotic_model_b_backend_name=emotic_model_b_backend_name,
                                          emotic_model_c_backend_name=emotic_model_c_backend_name,
                                          hra_model_backend_name=hra_model_backend_name,
                                          nb_of_fine_tuned_conv_layers=nb_of_conv_layers_to_fine_tune,
                                          violation_class=violation_class)

print (DisplaceNet_preds)
print (DisplaceNet_preds[0])


#!/usr/bin/env python

"""
Copyright 2017-2018 Fizyr (https://fizyr.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import argparse
import os
import sys

import keras
import tensorflow as tf

# Allow relative imports when being executed as script.
if __name__ == "__main__" and __package__ is None:
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..'))
    __package__ = "keras_retinanet.bin"

# Change these to absolute imports if you copy this script outside the keras_retinanet package.
from .. import models
from ..preprocessing.csv_generator import CSVGenerator
from ..preprocessing.pascal_voc import PascalVocGenerator
from ..utils.eval import evaluate
from ..utils.keras_version import check_keras_version


def get_session():
    """ Construct a modified tf session.
    """
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    return tf.Session(config=config)


def create_generator(args):
    """ Create generators for evaluation.
    """
    if args.dataset_type == 'coco':
        # import here to prevent unnecessary dependency on cocoapi
        from ..preprocessing.coco import CocoGenerator

        validation_generator = CocoGenerator(
            args.coco_path,
            'val2017',
            image_min_side=args.image_min_side,
            image_max_side=args.image_max_side
        )
    elif args.dataset_type == 'pascal':
        validation_generator = PascalVocGenerator(
            args.pascal_path,
            'test',
            image_min_side=args.image_min_side,
            image_max_side=args.image_max_side
        )
    elif args.dataset_type == 'csv':
        validation_generator = CSVGenerator(
            args.annotations,
            args.classes,
            image_min_side=args.image_min_side,
            image_max_side=args.image_max_side
        )
    else:
        raise ValueError('Invalid data type received: {}'.format(args.dataset_type))

    return validation_generator


def parse_args(args):
    """ Parse the arguments.
    """
    parser     = argparse.ArgumentParser(description='Evaluation script for a RetinaNet network.')
    subparsers = parser.add_subparsers(help='Arguments for specific dataset types.', dest='dataset_type')
    subparsers.required = True

    coco_parser = subparsers.add_parser('coco')
    coco_parser.add_argument('coco_path', help='Path to dataset directory (ie. /tmp/COCO).')

    pascal_parser = subparsers.add_parser('pascal')
    pascal_parser.add_argument('pascal_path', help='Path to dataset directory (ie. /tmp/VOCdevkit).')

    csv_parser = subparsers.add_parser('csv')
    csv_parser.add_argument('annotations', help='Path to CSV file containing annotations for evaluation.')
    csv_parser.add_argument('classes', help='Path to a CSV file containing class label mapping.')

    parser.add_argument('model',             help='Path to RetinaNet model.')
    parser.add_argument('--convert-model',   help='Convert the model to an inference model (ie. the input is a training model).', action='store_true')
    parser.add_argument('--backbone',        help='The backbone of the model.', default='resnet50')
    parser.add_argument('--gpu',             help='Id of the GPU to use (as reported by nvidia-smi).')
    parser.add_argument('--score-threshold', help='Threshold on score to filter detections with (defaults to 0.05).', default=0.05, type=float)
    parser.add_argument('--iou-threshold',   help='IoU Threshold to count for a positive detection (defaults to 0.5).', default=0.5, type=float)
    parser.add_argument('--max-detections',  help='Max Detections per image (defaults to 100).', default=100, type=int)
    parser.add_argument('--save-path',       help='Path for saving images with detections (doesn\'t work for COCO).')
    parser.add_argument('--image-min-side',  help='Rescale the image so the smallest side is min_side.', type=int, default=800)
    parser.add_argument('--image-max-side',  help='Rescale the image if the largest side is larger than max_side.', type=int, default=1333)

    return parser.parse_args(args)


def main(args=None):
    # parse arguments
    if args is None:
        args = sys.argv[1:]
    args = parse_args(args)

    # make sure keras is the minimum required version
    check_keras_version()

    # optionally choose specific GPU
    if args.gpu:
        os.environ['CUDA_VISIBLE_DEVICES'] = args.gpu
    keras.backend.tensorflow_backend.set_session(get_session())

    # make save path if it doesn't exist
    if args.save_path is not None and not os.path.exists(args.save_path):
        os.makedirs(args.save_path)

    # create the generator
    generator = create_generator(args)

    # load the model
    print('Loading model, this may take a second...')
    model = models.load_model(args.model, backbone_name=args.backbone, convert=args.convert_model)

    # print model summary
    # print(model.summary())

    # start evaluation
    if args.dataset_type == 'coco':
        from ..utils.coco_eval import evaluate_coco
        evaluate_coco(generator, model, args.score_threshold)
    else:
        average_precisions = evaluate(
            generator,
            model,
            iou_threshold=args.iou_threshold,
            score_threshold=args.score_threshold,
            max_detections=args.max_detections,
            save_path=args.save_path
        )

        # print evaluation
        for label, average_precision in average_precisions.items():
            print(generator.label_to_name(label), '{:.4f}'.format(average_precision))
        print('mAP: {:.4f}'.format(sum(average_precisions.values()) / len(average_precisions)))


if __name__ == '__main__':
    main()



"""Custom generator for pentuple-output Keras models.
"""

from math import ceil


def custom_generator(hdf5_file, nb_data, batch_size, mode):
    """ Generates batches of tensor image data in form of ==> (x1, y1) ,(x2, y2).
        # Reference
        - https://stackoverflow.com/questions/50333532/load-images-and-annotations-from-csv-and-use-fit-generator-with-multi-output-mod
        - http://machinelearninguru.com/deep_learning/data_preparation/hdf5/hdf5.html

        # Arguments
            hdf5_file: path or hdf5 object which contains the images and the annotations.
            nb_data: total number of samples saved in the array.
            batch_size: size of the batch to generate tensor image data for.
            module: one of `body` or `image`.

        # Returns
            A generator object.

    """

    batches_list = list(range(int(ceil(float(nb_data) / batch_size))))

    while True:

        # loop over batches
        for n, i in enumerate(batches_list):
            i_s = i * batch_size  # index of the first image in this batch
            i_e = min([(i + 1) * batch_size, nb_data])  # index of the last image in this batch

            if mode == 'train':
                body_x = hdf5_file["x_cropped_train"][i_s:i_e, ...]
                image_x = hdf5_file["x_entire_train"][i_s:i_e, ...]
                # valence_body_y = hdf5_file["valence_cropped_train"][i_s:i_e]
                valence_image_y = hdf5_file["valence_entire_train"][i_s:i_e]
                # arousal_body_y = hdf5_file["arousal_cropped_train"][i_s:i_e]
                arousal_image_y = hdf5_file["arousal_entire_train"][i_s:i_e]
                # dominance_body_y = hdf5_file["dominance_cropped_train"][i_s:i_e]
                dominance_image_y = hdf5_file["dominance_entire_train"][i_s:i_e]

            elif mode == 'val':
                body_x = hdf5_file["x_cropped_val"][i_s:i_e, ...]
                image_x = hdf5_file["x_entire_val"][i_s:i_e, ...]
                # valence_body_y = hdf5_file["valence_cropped_val"][i_s:i_e]
                valence_image_y = hdf5_file["valence_entire_val"][i_s:i_e]
                # arousal_body_y = hdf5_file["arousal_cropped_val"][i_s:i_e]
                arousal_image_y = hdf5_file["arousal_entire_val"][i_s:i_e]
                # dominance_body_y = hdf5_file["dominance_cropped_val"][i_s:i_e]
                dominance_image_y = hdf5_file["dominance_entire_val"][i_s:i_e]

            elif mode == 'test':
                body_x = hdf5_file["x_cropped_test"][i_s:i_e, ...]
                image_x = hdf5_file["x_entire_test"][i_s:i_e, ...]
                # valence_body_y = hdf5_file["valence_cropped_test"][i_s:i_e]
                valence_image_y = hdf5_file["valence_entire_test"][i_s:i_e]
                # arousal_body_y = hdf5_file["arousal_cropped_test"][i_s:i_e]
                arousal_image_y = hdf5_file["arousal_entire_test"][i_s:i_e]
                # dominance_body_y = hdf5_file["dominance_cropped_test"][i_s:i_e]
                dominance_image_y = hdf5_file["dominance_entire_test"][i_s:i_e]


        yield [body_x,image_x], [valence_image_y,arousal_image_y,dominance_image_y]

def custom_generator_single_output(hdf5_file, nb_data, batch_size, mode):
        """ Generates batches of tensor image data in form of ==> (x1, y1) ,(x2, y2).
            # Reference
            - https://stackoverflow.com/questions/50333532/load-images-and-annotations-from-csv-and-use-fit-generator-with-multi-output-mod
            - http://machinelearninguru.com/deep_learning/data_preparation/hdf5/hdf5.html

            # Arguments
                hdf5_file: path or hdf5 object which contains the images and the annotations.
                nb_data: total number of samples saved in the array.
                batch_size: size of the batch to generate tensor image data for.
                module: one of `body` or `image`.

            # Returns
                A generator object.

        """

        batches_list = list(range(int(ceil(float(nb_data) / batch_size))))

        while True:

            # loop over batches
            for n, i in enumerate(batches_list):
                i_s = i * batch_size  # index of the first image in this batch
                i_e = min([(i + 1) * batch_size, nb_data])  # index of the last image in this batch

                if mode == 'train':
                    body_x = hdf5_file["x_body_train"][i_s:i_e, ...]
                    image_x = hdf5_file["x_image_train"][i_s:i_e, ...]
                    y = hdf5_file["y_image_train"][i_s:i_e]

                elif mode == 'val':
                    body_x = hdf5_file["x_body_val"][i_s:i_e, ...]
                    image_x = hdf5_file["x_image_val"][i_s:i_e, ...]
                    y = hdf5_file["y_image_val"][i_s:i_e]

                elif mode == 'test':
                    body_x = hdf5_file["x_body_test"][i_s:i_e, ...]
                    image_x = hdf5_file["x_image_test"][i_s:i_e, ...]
                    y = hdf5_file["y_image_test"][i_s:i_e]

            yield [body_x, image_x], y


"""
Copyright 2017-2018 Fizyr (https://fizyr.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import random
import threading
import warnings

import keras

from ..utils.anchors import anchor_targets_bbox, bbox_transform
from ..utils.image import (
    TransformParameters,
    adjust_transform_for_image,
    apply_transform,
    preprocess_image,
    resize_image,
)
from ..utils.transform import transform_aabb


class Generator(object):
    """ Abstract generator class.
    """

    def __init__(
        self,
        transform_generator = None,
        batch_size=1,
        group_method='ratio',  # one of 'none', 'random', 'ratio'
        shuffle_groups=True,
        image_min_side=800,
        image_max_side=1333,
        transform_parameters=None,
        compute_anchor_targets=anchor_targets_bbox,
    ):
        """ Initialize Generator object.

        Args
            transform_generator    : A generator used to randomly transform images and annotations.
            batch_size             : The size of the batches to generate.
            group_method           : Determines how images are grouped together (defaults to 'ratio', one of ('none', 'random', 'ratio')).
            shuffle_groups         : If True, shuffles the groups each epoch.
            image_min_side         : After resizing the minimum side of an image is equal to image_min_side.
            image_max_side         : If after resizing the maximum side is larger than image_max_side, scales down further so that the max side is equal to image_max_side.
            transform_parameters   : The transform parameters used for data augmentation.
            compute_anchor_targets : Function handler for computing the targets of anchors for an image and its annotations.
        """
        self.transform_generator    = transform_generator
        self.batch_size             = int(batch_size)
        self.group_method           = group_method
        self.shuffle_groups         = shuffle_groups
        self.image_min_side         = image_min_side
        self.image_max_side         = image_max_side
        self.transform_parameters   = transform_parameters or TransformParameters()
        self.compute_anchor_targets = compute_anchor_targets

        self.group_index = 0
        self.lock        = threading.Lock()

        self.group_images()

    def size(self):
        """ Size of the dataset.
        """
        raise NotImplementedError('size method not implemented')

    def num_classes(self):
        """ Number of classes in the dataset.
        """
        raise NotImplementedError('num_classes method not implemented')

    def name_to_label(self, name):
        """ Map name to label.
        """
        raise NotImplementedError('name_to_label method not implemented')

    def label_to_name(self, label):
        """ Map label to name.
        """
        raise NotImplementedError('label_to_name method not implemented')

    def image_aspect_ratio(self, image_index):
        """ Compute the aspect ratio for an image with image_index.
        """
        raise NotImplementedError('image_aspect_ratio method not implemented')

    def load_image(self, image_index):
        """ Load an image at the image_index.
        """
        raise NotImplementedError('load_image method not implemented')

    def load_annotations(self, image_index):
        """ Load annotations for an image_index.
        """
        raise NotImplementedError('load_annotations method not implemented')

    def load_annotations_group(self, group):
        """ Load annotations for all images in group.
        """
        return [self.load_annotations(image_index) for image_index in group]

    def filter_annotations(self, image_group, annotations_group, group):
        """ Filter annotations by removing those that are outside of the image bounds or whose width/height < 0.
        """
        # test all annotations
        for index, (image, annotations) in enumerate(zip(image_group, annotations_group)):
            assert(isinstance(annotations, np.ndarray)), '\'load_annotations\' should return a list of numpy arrays, received: {}'.format(type(annotations))

            # test x2 < x1 | y2 < y1 | x1 < 0 | y1 < 0 | x2 <= 0 | y2 <= 0 | x2 >= image.shape[1] | y2 >= image.shape[0]
            invalid_indices = np.where(
                (annotations[:, 2] <= annotations[:, 0]) |
                (annotations[:, 3] <= annotations[:, 1]) |
                (annotations[:, 0] < 0) |
                (annotations[:, 1] < 0) |
                (annotations[:, 2] > image.shape[1]) |
                (annotations[:, 3] > image.shape[0])
            )[0]

            # delete invalid indices
            if len(invalid_indices):
                warnings.warn('Image with id {} (shape {}) contains the following invalid boxes: {}.'.format(
                    group[index],
                    image.shape,
                    [annotations[invalid_index, :] for invalid_index in invalid_indices]
                ))
                annotations_group[index] = np.delete(annotations, invalid_indices, axis=0)

        return image_group, annotations_group

    def load_image_group(self, group):
        """ Load images for all images in a group.
        """
        return [self.load_image(image_index) for image_index in group]

    def random_transform_group_entry(self, image, annotations):
        """ Randomly transforms image and annotation.
        """
        # randomly transform both image and annotations
        if self.transform_generator:
            transform = adjust_transform_for_image(next(self.transform_generator), image, self.transform_parameters.relative_translation)
            image     = apply_transform(transform, image, self.transform_parameters)

            # Transform the bounding boxes in the annotations.
            annotations = annotations.copy()
            for index in range(annotations.shape[0]):
                annotations[index, :4] = transform_aabb(transform, annotations[index, :4])

        return image, annotations

    def resize_image(self, image):
        """ Resize an image using image_min_side and image_max_side.
        """
        return resize_image(image, min_side=self.image_min_side, max_side=self.image_max_side)

    def preprocess_image(self, image):
        """ Preprocess an image (e.g. subtracts ImageNet mean).
        """
        return preprocess_image(image)

    def preprocess_group_entry(self, image, annotations):
        """ Preprocess image and its annotations.
        """
        # preprocess the image
        image = self.preprocess_image(image)

        # randomly transform image and annotations
        image, annotations = self.random_transform_group_entry(image, annotations)

        # resize image
        image, image_scale = self.resize_image(image)

        # apply resizing to annotations too
        annotations[:, :4] *= image_scale

        return image, annotations

    def preprocess_group(self, image_group, annotations_group):
        """ Preprocess each image and its annotations in its group.
        """
        for index, (image, annotations) in enumerate(zip(image_group, annotations_group)):
            # preprocess a single group entry
            image, annotations = self.preprocess_group_entry(image, annotations)

            # copy processed data back to group
            image_group[index]       = image
            annotations_group[index] = annotations

        return image_group, annotations_group

    def group_images(self):
        """ Order the images according to self.order and makes groups of self.batch_size.
        """
        # determine the order of the images
        order = list(range(self.size()))
        if self.group_method == 'random':
            random.shuffle(order)
        elif self.group_method == 'ratio':
            order.sort(key=lambda x: self.image_aspect_ratio(x))

        # divide into groups, one group = one batch
        self.groups = [[order[x % len(order)] for x in range(i, i + self.batch_size)] for i in range(0, len(order), self.batch_size)]

    def compute_inputs(self, image_group):
        """ Compute inputs for the network using an image_group.
        """
        # get the max image shape
        max_shape = tuple(max(image.shape[x] for image in image_group) for x in range(3))

        # construct an image batch object
        image_batch = np.zeros((self.batch_size,) + max_shape, dtype=keras.backend.floatx())

        # copy all images to the upper left part of the image batch object
        for image_index, image in enumerate(image_group):
            image_batch[image_index, :image.shape[0], :image.shape[1], :image.shape[2]] = image

        return image_batch

    def compute_targets(self, image_group, annotations_group):
        """ Compute target outputs for the network using images and their annotations.
        """
        # get the max image shape
        max_shape = tuple(max(image.shape[x] for image in image_group) for x in range(3))

        # compute labels and regression targets
        labels_group     = [None] * self.batch_size
        regression_group = [None] * self.batch_size
        for index, (image, annotations) in enumerate(zip(image_group, annotations_group)):
            # compute regression targets
            labels_group[index], annotations, anchors = self.compute_anchor_targets(
                max_shape,
                annotations,
                self.num_classes(),
                mask_shape=image.shape,
            )
            regression_group[index] = bbox_transform(anchors, annotations)

            # append anchor states to regression targets (necessary for filtering 'ignore', 'positive' and 'negative' anchors)
            anchor_states           = np.max(labels_group[index], axis=1, keepdims=True)
            regression_group[index] = np.append(regression_group[index], anchor_states, axis=1)

        labels_batch     = np.zeros((self.batch_size,) + labels_group[0].shape, dtype=keras.backend.floatx())
        regression_batch = np.zeros((self.batch_size,) + regression_group[0].shape, dtype=keras.backend.floatx())

        # copy all labels and regression values to the batch blob
        for index, (labels, regression) in enumerate(zip(labels_group, regression_group)):
            labels_batch[index, ...]     = labels
            regression_batch[index, ...] = regression

        return [regression_batch, labels_batch]

    def compute_input_output(self, group):
        """ Compute inputs and target outputs for the network.
        """
        # load images and annotations
        image_group       = self.load_image_group(group)
        annotations_group = self.load_annotations_group(group)

        # check validity of annotations
        image_group, annotations_group = self.filter_annotations(image_group, annotations_group, group)

        # perform preprocessing steps
        image_group, annotations_group = self.preprocess_group(image_group, annotations_group)

        # compute network inputs
        inputs = self.compute_inputs(image_group)

        # compute network targets
        targets = self.compute_targets(image_group, annotations_group)

        return inputs, targets

    def __next__(self):
        return self.next()

    def next(self):
        # advance the group index
        with self.lock:
            if self.group_index == 0 and self.shuffle_groups:
                # shuffle groups at start of epoch
                random.shuffle(self.groups)
            group = self.groups[self.group_index]
            self.group_index = (self.group_index + 1) % len(self.groups)

        return self.compute_input_output(group)



"""
Copyright 2017-2018 Fizyr (https://fizyr.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np

import keras
import keras_retinanet.layers


class TestFilterDetections(object):
    def test_simple(self):
        # create simple FilterDetections layer
        filter_detections_layer = keras_retinanet.layers.FilterDetections()

        # create simple input
        boxes = np.array([[
            [0, 0, 10, 10],
            [0, 0, 10, 10],  # this will be suppressed
        ]], dtype=keras.backend.floatx())
        boxes = keras.backend.variable(boxes)

        classification = np.array([[
            [0, 0.9],  # this will be suppressed
            [0, 1],
        ]], dtype=keras.backend.floatx())
        classification = keras.backend.variable(classification)

        # compute output
        actual_boxes, actual_scores, actual_labels = filter_detections_layer.call([boxes, classification])
        actual_boxes  = keras.backend.eval(actual_boxes)
        actual_scores = keras.backend.eval(actual_scores)
        actual_labels = keras.backend.eval(actual_labels)

        # define expected output
        expected_boxes = -1 * np.ones((1, 300, 4), dtype=keras.backend.floatx())
        expected_boxes[0, 0, :] = [0, 0, 10, 10]

        expected_scores = -1 * np.ones((1, 300), dtype=keras.backend.floatx())
        expected_scores[0, 0] = 1

        expected_labels = -1 * np.ones((1, 300), dtype=keras.backend.floatx())
        expected_labels[0, 0] = 1

        # assert actual and expected are equal
        np.testing.assert_array_equal(actual_boxes, expected_boxes)
        np.testing.assert_array_equal(actual_scores, expected_scores)
        np.testing.assert_array_equal(actual_labels, expected_labels)

    def test_simple_with_other(self):
        # create simple FilterDetections layer
        filter_detections_layer = keras_retinanet.layers.FilterDetections()

        # create simple input
        boxes = np.array([[
            [0, 0, 10, 10],
            [0, 0, 10, 10],  # this will be suppressed
        ]], dtype=keras.backend.floatx())
        boxes = keras.backend.variable(boxes)

        classification = np.array([[
            [0, 0.9],  # this will be suppressed
            [0, 1],
        ]], dtype=keras.backend.floatx())
        classification = keras.backend.variable(classification)

        other = []
        other.append(np.array([[
            [0, 1234],  # this will be suppressed
            [0, 5678],
        ]], dtype=keras.backend.floatx()))
        other.append(np.array([[
            5678,  # this will be suppressed
            1234,
        ]], dtype=keras.backend.floatx()))
        other = [keras.backend.variable(o) for o in other]

        # compute output
        actual = filter_detections_layer.call([boxes, classification] + other)
        actual_boxes  = keras.backend.eval(actual[0])
        actual_scores = keras.backend.eval(actual[1])
        actual_labels = keras.backend.eval(actual[2])
        actual_other  = [keras.backend.eval(a) for a in actual[3:]]

        # define expected output
        expected_boxes = -1 * np.ones((1, 300, 4), dtype=keras.backend.floatx())
        expected_boxes[0, 0, :] = [0, 0, 10, 10]

        expected_scores = -1 * np.ones((1, 300), dtype=keras.backend.floatx())
        expected_scores[0, 0] = 1

        expected_labels = -1 * np.ones((1, 300), dtype=keras.backend.floatx())
        expected_labels[0, 0] = 1

        expected_other = []
        expected_other.append(-1 * np.ones((1, 300, 2), dtype=keras.backend.floatx()))
        expected_other[-1][0, 0, :] = [0, 5678]
        expected_other.append(-1 * np.ones((1, 300), dtype=keras.backend.floatx()))
        expected_other[-1][0, 0] = 1234

        # assert actual and expected are equal
        np.testing.assert_array_equal(actual_boxes, expected_boxes)
        np.testing.assert_array_equal(actual_scores, expected_scores)
        np.testing.assert_array_equal(actual_labels, expected_labels)

        for a, e in zip(actual_other, expected_other):
            np.testing.assert_array_equal(a, e)

    def test_mini_batch(self):
        # create simple FilterDetections layer
        filter_detections_layer = keras_retinanet.layers.FilterDetections()

        # create input with batch_size=2
        boxes = np.array([
            [
                [0, 0, 10, 10],  # this will be suppressed
                [0, 0, 10, 10],
            ],
            [
                [100, 100, 150, 150],
                [100, 100, 150, 150],  # this will be suppressed
            ],
        ], dtype=keras.backend.floatx())
        boxes = keras.backend.variable(boxes)

        classification = np.array([
            [
                [0, 0.9],  # this will be suppressed
                [0, 1],
            ],
            [
                [1,   0],
                [0.9, 0],  # this will be suppressed
            ],
        ], dtype=keras.backend.floatx())
        classification = keras.backend.variable(classification)

        # compute output
        actual_boxes, actual_scores, actual_labels = filter_detections_layer.call([boxes, classification])
        actual_boxes  = keras.backend.eval(actual_boxes)
        actual_scores = keras.backend.eval(actual_scores)
        actual_labels = keras.backend.eval(actual_labels)

        # define expected output
        expected_boxes = -1 * np.ones((2, 300, 4), dtype=keras.backend.floatx())
        expected_boxes[0, 0, :] = [0, 0, 10, 10]
        expected_boxes[1, 0, :] = [100, 100, 150, 150]

        expected_scores = -1 * np.ones((2, 300), dtype=keras.backend.floatx())
        expected_scores[0, 0] = 1
        expected_scores[1, 0] = 1

        expected_labels = -1 * np.ones((2, 300), dtype=keras.backend.floatx())
        expected_labels[0, 0] = 1
        expected_labels[1, 0] = 0

        # assert actual and expected are equal
        np.testing.assert_array_equal(actual_boxes, expected_boxes)
        np.testing.assert_array_equal(actual_scores, expected_scores)
        np.testing.assert_array_equal(actual_labels, expected_labels)



import keras.callbacks


class RedirectModel(keras.callbacks.Callback):
    """Callback which wraps another callback, but executed on a different model.

    ```python
    model = keras.models.load_model('model.h5')
    model_checkpoint = ModelCheckpoint(filepath='snapshot.h5')
    parallel_model = multi_gpu_model(model, gpus=2)
    parallel_model.fit(X_train, Y_train, callbacks=[RedirectModel(model_checkpoint, model)])
    ```

    Args
        callback : callback to wrap.
        model    : model to use when executing callbacks.
    """

    def __init__(self,
                 callback,
                 model):
        super(RedirectModel, self).__init__()

        self.callback = callback
        self.redirect_model = model

    def on_epoch_begin(self, epoch, logs=None):
        self.callback.on_epoch_begin(epoch, logs=logs)

    def on_epoch_end(self, epoch, logs=None):
        self.callback.on_epoch_end(epoch, logs=logs)

    def on_batch_begin(self, batch, logs=None):
        self.callback.on_batch_begin(batch, logs=logs)

    def on_batch_end(self, batch, logs=None):
        self.callback.on_batch_end(batch, logs=logs)

    def on_train_begin(self, logs=None):
        # overwrite the model with our custom model
        self.callback.set_model(self.redirect_model)

        self.callback.on_train_begin(logs=logs)

    def on_train_end(self, logs=None):
        self.callback.on_train_end(logs=logs)



# -*- coding: utf-8 -*-
'''


'''

from engine.displaced_people_branch.single_image_inference_hra_2class import single_img_HRA_inference
from applications.hra_utils import plot_preds
from keras.preprocessing import image
from matplotlib import pyplot as plt
from scipy.misc import imread


img_path = '/home/sandbox/Desktop/Testing Images/camping.jpg'
violation_class = 'dp'
model_backend_name = 'VGG16'
nb_of_conv_layers_to_fine_tune = 1

raw_preds, overlayed_text = single_img_HRA_inference(img_path=img_path,
                                     violation_class=violation_class,
                                     model_backend_name=model_backend_name,
                                     nb_of_conv_layers_to_fine_tune=nb_of_conv_layers_to_fine_tune)


img = image.load_img(img_path, target_size=(224, 224))
print (raw_preds[0])

# plot_preds(violation_class, img, raw_preds[0])


numpy_img_path = imread(img_path)
plt.figure(figsize=(10, 12))
plt.imshow(numpy_img_path)

current_axis = plt.gca()

# configure colours for bounding box and text
bounding_box_colour_rgbvar = (53, 42, 146)
bounding_box_colour_rgbvar2 = [x / 255.0 for x in bounding_box_colour_rgbvar]

text_colour_rgbvar = (214, 86, 100)
text_colour_rgbvar2 = [x / 255.0 for x in text_colour_rgbvar]

current_axis.text(0, 0, overlayed_text, size='x-large', color='white',
                  bbox={'facecolor': bounding_box_colour_rgbvar2, 'alpha': 1.0})

plt.axis('off')
plt.show()


"""Python utilities required to load data (image & their annotations) stored in numpy arrays.
    Functions `load_numpy_arrays_single_output` & `load_numpy_arrays_emotions_age_only` are deprecated.
    Use either the main function `load_data_from_numpy` to load all the applicable arrays
    or the supporting `load_annotations_only_from_numpy` instead.
"""

from __future__ import print_function
import numpy as np

def load_data_from_numpy(main_numpy_dir,
                         verbose = 1):

    print ('[INFO] Loading data from numpy arrays...')

    x_entire_train = np.load(main_numpy_dir + 'X_train/x_entire_train.npy')
    x_cropped_train = np.load(main_numpy_dir + 'X_train/x_cropped_train.npy')

    valence_entire_train = np.load(main_numpy_dir + 'Y_train/valence_train.npy')
    valence_cropped_train = np.load(main_numpy_dir + 'Y_train/valence_train.npy')

    arousal_entire_train = np.load(main_numpy_dir + 'Y_train/arousal_train.npy')
    arousal_cropped_train = np.load(main_numpy_dir + 'Y_train/arousal_train.npy')

    dominance_entire_train = np.load(main_numpy_dir + 'Y_train/dominance_train.npy')
    dominance_cropped_train = np.load(main_numpy_dir + 'Y_train/dominance_train.npy')



    x_entire_val = np.load(main_numpy_dir + 'X_train/x_entire_val.npy')
    x_cropped_val = np.load(main_numpy_dir + 'X_train/x_cropped_val.npy')



    valence_entire_val = np.load(main_numpy_dir + 'Y_train/valence_val.npy')
    valence_cropped_val = np.load(main_numpy_dir + 'Y_train/valence_val.npy')

    arousal_entire_val = np.load(main_numpy_dir + 'Y_train/arousal_val.npy')
    arousal_cropped_val = np.load(main_numpy_dir + 'Y_train/arousal_val.npy')

    dominance_entire_val = np.load(main_numpy_dir + 'Y_train/dominance_val.npy')
    dominance_cropped_val = np.load(main_numpy_dir + 'Y_train/dominance_val.npy')



    x_entire_test = np.load(main_numpy_dir + 'X_train/x_entire_test.npy')
    x_cropped_test = np.load(main_numpy_dir + 'X_train/x_cropped_test.npy')

    valence_entire_test = np.load(main_numpy_dir + 'Y_train/valence_test.npy')
    valence_cropped_test = np.load(main_numpy_dir + 'Y_train/valence_test.npy')

    arousal_entire_test = np.load(main_numpy_dir + 'Y_train/arousal_test.npy')
    arousal_cropped_test = np.load(main_numpy_dir + 'Y_train/arousal_test.npy')

    dominance_entire_test = np.load(main_numpy_dir + 'Y_train/dominance_test.npy')
    dominance_cropped_test = np.load(main_numpy_dir + 'Y_train/dominance_test.npy')



    print('[INFO] Data have been successfully loaded')
    print('---------------------------------------------------------------------------------------------------')
    if verbose == 1:
        print('x_entire_train shape:', x_entire_train.shape)
        print('x_cropped_train shape:', x_cropped_train.shape)
        print('valence_entire_train shape:', valence_entire_train.shape)
        print('valence_cropped_train shape:', valence_cropped_train.shape)
        print('arousal_entire_train shape:', arousal_entire_train.shape)
        print('arousal_cropped_train shape:', arousal_cropped_train.shape)
        print('dominance_entire_train shape:', dominance_entire_train.shape)
        print('dominance_cropped_train shape:', dominance_cropped_train.shape)

        print ('---------------------------------------------------------------------------------------------------')

        print('x_entire_val shape:', x_entire_val.shape)
        print('x_cropped_val shape:', x_cropped_val.shape)
        print('valence_entire_val shape:', valence_entire_val.shape)
        print('valence_cropped_val shape:', valence_cropped_val.shape)
        print('arousal_entire_val shape:', arousal_entire_val.shape)
        print('arousal_cropped_val shape:', arousal_cropped_val.shape)
        print('dominance_entire_val shape:', dominance_entire_val.shape)
        print('dominance_cropped_val shape:', dominance_cropped_val.shape)

        print ('---------------------------------------------------------------------------------------------------')

        print('x_entire_test shape:', x_entire_test.shape)
        print('x_cropped_test shape:', x_cropped_test.shape)
        print('valence_entire_test shape:', valence_entire_test.shape)
        print('valence_cropped_test shape:', valence_cropped_test.shape)
        print('arousal_entire_test shape:', arousal_entire_test.shape)
        print('arousal_cropped_test shape:', arousal_cropped_test.shape)
        print('dominance_entire_test shape:', dominance_entire_test.shape)
        print('dominance_cropped_test shape:', dominance_cropped_test.shape)

        print ('---------------------------------------------------------------------------------------------------')

    return (x_entire_train, x_cropped_train,valence_entire_train,valence_cropped_train,arousal_entire_train,arousal_cropped_train,dominance_entire_train,dominance_cropped_train), \
           (x_entire_val, x_cropped_val, valence_entire_val,valence_cropped_val,arousal_entire_val,arousal_cropped_val,dominance_entire_val,dominance_cropped_val), \
           (x_entire_test, x_cropped_test, valence_entire_test,valence_cropped_test,arousal_entire_test,arousal_cropped_test,dominance_entire_test,dominance_cropped_test)



def load_data_from_numpy_single_output(main_numpy_dir,
                                       verbose=1):

    print ('[INFO] Loading data from numpy arrays...')

    x_image_train = np.load(main_numpy_dir + 'X_train/x_image_train.npy')
    x_body_train = np.load(main_numpy_dir + 'X_train/x_body_train.npy')

    y_image_train = np.load(main_numpy_dir + 'Y_train/y_train.npy')
    y_body_train = np.load(main_numpy_dir + 'Y_train/y_train.npy')


    x_image_val = np.load(main_numpy_dir + 'X_train/x_image_val.npy')
    x_body_val = np.load(main_numpy_dir + 'X_train/x_body_val.npy')

    y_image_val = np.load(main_numpy_dir + 'Y_train/y_val.npy')
    y_body_val = np.load(main_numpy_dir + 'Y_train/y_val.npy')


    x_image_test = np.load(main_numpy_dir + 'X_train/x_image_test.npy')
    x_body_test = np.load(main_numpy_dir + 'X_train/x_body_test.npy')

    y_image_test = np.load(main_numpy_dir + 'Y_train/y_test.npy')
    y_body_test = np.load(main_numpy_dir + 'Y_train/y_test.npy')



    print('[INFO] Data have been successfully loaded')
    print('---------------------------------------------------------------------------------------------------')
    if verbose == 1:
        print('x_image_train shape:', x_image_train.shape)
        print('x_body_train shape:', x_body_train.shape)
        print('y_image_train shape:', y_image_train.shape)
        print('y_body_train shape:', y_body_train.shape)


        print ('---------------------------------------------------------------------------------------------------')

        print('x_image_val shape:', x_image_val.shape)
        print('x_body_val shape:', x_body_val.shape)
        print('y_image_val shape:', y_image_val.shape)
        print('y_body_val shape:', y_body_val.shape)


        print ('---------------------------------------------------------------------------------------------------')

        print('x_image_test shape:', x_image_test.shape)
        print('x_body_test shape:', x_body_test.shape)
        print('y_image_test shape:', y_image_test.shape)
        print('y_body_test shape:', y_body_test.shape)


        print ('---------------------------------------------------------------------------------------------------')

    return (x_image_train, x_body_train,y_image_train,y_body_train), \
           (x_image_val, x_body_val, y_image_val,y_body_val), \
           (x_image_test, x_body_test,y_image_test,y_body_test)










'''
Utilities that are useful to sub- or up-sample weights tensors.

Copyright (C) 2018 Pierluigi Ferrari

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

import numpy as np

def sample_tensors(weights_list, sampling_instructions, axes=None, init=None, mean=0.0, stddev=0.005):
    '''
    Can sub-sample and/or up-sample individual dimensions of the tensors in the given list
    of input tensors.

    It is possible to sub-sample some dimensions and up-sample other dimensions at the same time.

    The tensors in the list will be sampled consistently, i.e. for any given dimension that
    corresponds among all tensors in the list, the same elements will be picked for every tensor
    along that dimension.

    For dimensions that are being sub-sampled, you can either provide a list of the indices
    that should be picked, or you can provide the number of elements to be sub-sampled, in which
    case the elements will be chosen at random.

    For dimensions that are being up-sampled, "filler" elements will be insterted at random
    positions along the respective dimension. These filler elements will be initialized either
    with zero or from a normal distribution with selectable mean and standard deviation.

    Arguments:
        weights_list (list): A list of Numpy arrays. Each array represents one of the tensors
            to be sampled. The tensor with the greatest number of dimensions must be the first
            element in the list. For example, in the case of the weights of a 2D convolutional
            layer, the kernel must be the first element in the list and the bias the second,
            not the other way around. For all tensors in the list after the first tensor, the
            lengths of each of their axes must identical to the length of some axis of the
            first tensor.
        sampling_instructions (list): A list that contains the sampling instructions for each
            dimension of the first tensor. If the first tensor has `n` dimensions, then this
            must be a list of length `n`. That means, sampling instructions for every dimension
            of the first tensor must still be given even if not all dimensions should be changed.
            The elements of this list can be either lists of integers or integers. If the sampling
            instruction for a given dimension is a list of integers, then these integers represent
            the indices of the elements of that dimension that will be sub-sampled. If the sampling
            instruction for a given dimension is an integer, then that number of elements will be
            sampled along said dimension. If the integer is greater than the number of elements
            of the input tensors in that dimension, that dimension will be up-sampled. If the integer
            is smaller than the number of elements of the input tensors in that dimension, that
            dimension will be sub-sampled. If the integer is equal to the number of elements
            of the input tensors in that dimension, that dimension will remain the same.
        axes (list, optional): Only relevant if `weights_list` contains more than one tensor.
            This list contains a list for each additional tensor in `weights_list` beyond the first.
            Each of these lists contains integers that determine to which axes of the first tensor
            the axes of the respective tensor correspond. For example, let the first tensor be a
            4D tensor and the second tensor in the list be a 2D tensor. If the first element of
            `axis` is the list `[2,3]`, then that means that the two axes of the second tensor
            correspond to the last two axes of the first tensor, in the same order. The point of
            this list is for the program to know, if a given dimension of the first tensor is to
            be sub- or up-sampled, which dimensions of the other tensors in the list must be
            sub- or up-sampled accordingly.
        init (list, optional): Only relevant for up-sampling. Must be `None` or a list of strings
            that determines for each tensor in `weights_list` how the newly inserted values should
            be initialized. The possible values are 'gaussian' for initialization from a normal
            distribution with the selected mean and standard deviation (see the following two arguments),
            or 'zeros' for zero-initialization. If `None`, all initializations default to
            'gaussian'.
        mean (float, optional): Only relevant for up-sampling. The mean of the values that will
            be inserted into the tensors at random in the case of up-sampling.
        stddev (float, optional): Only relevant for up-sampling. The standard deviation of the
            values that will be inserted into the tensors at random in the case of up-sampling.

    Returns:
        A list containing the sampled tensors in the same order in which they were given.
    '''

    first_tensor = weights_list[0]

    if (not isinstance(sampling_instructions, (list, tuple))) or (len(sampling_instructions) != first_tensor.ndim):
        raise ValueError("The sampling instructions must be a list whose length is the number of dimensions of the first tensor in `weights_list`.")

    if (not init is None) and len(init) != len(weights_list):
        raise ValueError("`init` must either be `None` or a list of strings that has the same length as `weights_list`.")

    up_sample = [] # Store the dimensions along which we need to up-sample.
    out_shape = [] # Store the shape of the output tensor here.
    # Store two stages of the new (sub-sampled and/or up-sampled) weights tensors in the following two lists.
    subsampled_weights_list = [] # Tensors after sub-sampling, but before up-sampling (if any).
    upsampled_weights_list = [] # Sub-sampled tensors after up-sampling (if any), i.e. final output tensors.

    # Create the slicing arrays from the sampling instructions.
    sampling_slices = []
    for i, sampling_inst in enumerate(sampling_instructions):
        if isinstance(sampling_inst, (list, tuple)):
            amax = np.amax(np.array(sampling_inst))
            if amax >= first_tensor.shape[i]:
                raise ValueError("The sample instructions for dimension {} contain index {}, which is greater than the length of that dimension.".format(i, amax))
            sampling_slices.append(np.array(sampling_inst))
            out_shape.append(len(sampling_inst))
        elif isinstance(sampling_inst, int):
            out_shape.append(sampling_inst)
            if sampling_inst == first_tensor.shape[i]:
                # Nothing to sample here, we're keeping the original number of elements along this axis.
                sampling_slice = np.arange(sampling_inst)
                sampling_slices.append(sampling_slice)
            elif sampling_inst < first_tensor.shape[i]:
                # We want to SUB-sample this dimension. Randomly pick `sample_inst` many elements from it.
                sampling_slice1 = np.array([0]) # We will always sample class 0, the background class.
                # Sample the rest of the classes.
                sampling_slice2 = np.sort(np.random.choice(np.arange(1, first_tensor.shape[i]), sampling_inst - 1, replace=False))
                sampling_slice = np.concatenate([sampling_slice1, sampling_slice2])
                sampling_slices.append(sampling_slice)
            else:
                # We want to UP-sample. Pick all elements from this dimension.
                sampling_slice = np.arange(first_tensor.shape[i])
                sampling_slices.append(sampling_slice)
                up_sample.append(i)
        else:
            raise ValueError("Each element of the sampling instructions must be either an integer or a list/tuple of integers, but received `{}`".format(type(sampling_inst)))

    # Process the first tensor.
    subsampled_first_tensor = np.copy(first_tensor[np.ix_(*sampling_slices)])
    subsampled_weights_list.append(subsampled_first_tensor)

    # Process the other tensors.
    if len(weights_list) > 1:
        for j in range(1, len(weights_list)):
            this_sampling_slices = [sampling_slices[i] for i in axes[j-1]] # Get the sampling slices for this tensor.
            subsampled_weights_list.append(np.copy(weights_list[j][np.ix_(*this_sampling_slices)]))

    if up_sample:
        # Take care of the dimensions that are to be up-sampled.

        out_shape = np.array(out_shape)

        # Process the first tensor.
        if init is None or init[0] == 'gaussian':
            upsampled_first_tensor = np.random.normal(loc=mean, scale=stddev, size=out_shape)
        elif init[0] == 'zeros':
            upsampled_first_tensor = np.zeros(out_shape)
        else:
            raise ValueError("Valid initializations are 'gaussian' and 'zeros', but received '{}'.".format(init[0]))
        # Pick the indices of the elements in `upsampled_first_tensor` that should be occupied by `subsampled_first_tensor`.
        up_sample_slices = [np.arange(k) for k in subsampled_first_tensor.shape]
        for i in up_sample:
            # Randomly select across which indices of this dimension to scatter the elements of `new_weights_tensor` in this dimension.
            up_sample_slice1 = np.array([0])
            up_sample_slice2 = np.sort(np.random.choice(np.arange(1, upsampled_first_tensor.shape[i]), subsampled_first_tensor.shape[i] - 1, replace=False))
            up_sample_slices[i] = np.concatenate([up_sample_slice1, up_sample_slice2])
        upsampled_first_tensor[np.ix_(*up_sample_slices)] = subsampled_first_tensor
        upsampled_weights_list.append(upsampled_first_tensor)

        # Process the other tensors
        if len(weights_list) > 1:
            for j in range(1, len(weights_list)):
                if init is None or init[j] == 'gaussian':
                    upsampled_tensor = np.random.normal(loc=mean, scale=stddev, size=out_shape[axes[j-1]])
                elif init[j] == 'zeros':
                    upsampled_tensor = np.zeros(out_shape[axes[j-1]])
                else:
                    raise ValueError("Valid initializations are 'gaussian' and 'zeros', but received '{}'.".format(init[j]))
                this_up_sample_slices = [up_sample_slices[i] for i in axes[j-1]] # Get the up-sampling slices for this tensor.
                upsampled_tensor[np.ix_(*this_up_sample_slices)] = subsampled_weights_list[j]
                upsampled_weights_list.append(upsampled_tensor)

        return upsampled_weights_list
    else:
        return subsampled_weights_list



'''
Various geometric image transformations for 2D object detection, both deterministic
and probabilistic.

Copyright (C) 2018 Pierluigi Ferrari

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from __future__ import division

import numpy as np
import random

import cv2

from engine.object_detection_branch.single_shot_detector.data_generator.object_detection_2d_image_boxes_validation_utils import BoxFilter, ImageValidator


class Resize:
    '''
    Resizes images to a specified height and width in pixels.
    '''

    def __init__(self,
                 height,
                 width,
                 interpolation_mode=cv2.INTER_LINEAR,
                 box_filter=None,
                 labels_format={'class_id': 0, 'xmin': 1, 'ymin': 2, 'xmax': 3, 'ymax': 4}):
        '''
        Arguments:
            height (int): The desired height of the output images in pixels.
            width (int): The desired width of the output images in pixels.
            interpolation_mode (int, optional): An integer that denotes a valid
                OpenCV interpolation mode. For example, integers 0 through 5 are
                valid interpolation modes.
            box_filter (BoxFilter, optional): Only relevant if ground truth bounding boxes are given.
                A `BoxFilter` object to filter out bounding boxes that don't meet the given criteria
                after the transformation. Refer to the `BoxFilter` documentation for details. If `None`,
                the validity of the bounding boxes is not checked.
            labels_format (dict, optional): A dictionary that defines which index in the last axis of the labels
                of an image contains which bounding box coordinate. The dictionary maps at least the keywords
                'xmin', 'ymin', 'xmax', and 'ymax' to their respective indices within last axis of the labels array.
        '''
        if not (isinstance(box_filter, BoxFilter) or box_filter is None):
            raise ValueError("`box_filter` must be either `None` or a `BoxFilter` object.")
        self.out_height = height
        self.out_width = width
        self.interpolation_mode = interpolation_mode
        self.box_filter = box_filter
        self.labels_format = labels_format

    def __call__(self, image, labels=None, return_inverter=False):

        img_height, img_width = image.shape[:2]

        xmin = self.labels_format['xmin']
        ymin = self.labels_format['ymin']
        xmax = self.labels_format['xmax']
        ymax = self.labels_format['ymax']

        image = cv2.resize(image,
                           dsize=(self.out_width, self.out_height),
                           interpolation=self.interpolation_mode)

        if return_inverter:
            def inverter(labels):
                labels = np.copy(labels)
                labels[:, [ymin+1, ymax+1]] = np.round(labels[:, [ymin+1, ymax+1]] * (img_height / self.out_height), decimals=0)
                labels[:, [xmin+1, xmax+1]] = np.round(labels[:, [xmin+1, xmax+1]] * (img_width / self.out_width), decimals=0)
                return labels

        if labels is None:
            if return_inverter:
                return image, inverter
            else:
                return image
        else:
            labels = np.copy(labels)
            labels[:, [ymin, ymax]] = np.round(labels[:, [ymin, ymax]] * (self.out_height / img_height), decimals=0)
            labels[:, [xmin, xmax]] = np.round(labels[:, [xmin, xmax]] * (self.out_width / img_width), decimals=0)

            if not (self.box_filter is None):
                self.box_filter.labels_format = self.labels_format
                labels = self.box_filter(labels=labels,
                                         image_height=self.out_height,
                                         image_width=self.out_width)

            if return_inverter:
                return image, labels, inverter
            else:
                return image, labels

class ResizeRandomInterp:
    '''
    Resizes images to a specified height and width in pixels using a radnomly
    selected interpolation mode.
    '''

    def __init__(self,
                 height,
                 width,
                 interpolation_modes=[cv2.INTER_NEAREST,
                                      cv2.INTER_LINEAR,
                                      cv2.INTER_CUBIC,
                                      cv2.INTER_AREA,
                                      cv2.INTER_LANCZOS4],
                 box_filter=None,
                 labels_format={'class_id': 0, 'xmin': 1, 'ymin': 2, 'xmax': 3, 'ymax': 4}):
        '''
        Arguments:
            height (int): The desired height of the output image in pixels.
            width (int): The desired width of the output image in pixels.
            interpolation_modes (list/tuple, optional): A list/tuple of integers
                that represent valid OpenCV interpolation modes. For example,
                integers 0 through 5 are valid interpolation modes.
            box_filter (BoxFilter, optional): Only relevant if ground truth bounding boxes are given.
                A `BoxFilter` object to filter out bounding boxes that don't meet the given criteria
                after the transformation. Refer to the `BoxFilter` documentation for details. If `None`,
                the validity of the bounding boxes is not checked.
            labels_format (dict, optional): A dictionary that defines which index in the last axis of the labels
                of an image contains which bounding box coordinate. The dictionary maps at least the keywords
                'xmin', 'ymin', 'xmax', and 'ymax' to their respective indices within last axis of the labels array.
        '''
        if not (isinstance(interpolation_modes, (list, tuple))):
            raise ValueError("`interpolation_mode` must be a list or tuple.")
        self.height = height
        self.width = width
        self.interpolation_modes = interpolation_modes
        self.box_filter = box_filter
        self.labels_format = labels_format
        self.resize = Resize(height=self.height,
                             width=self.width,
                             box_filter=self.box_filter,
                             labels_format=self.labels_format)

    def __call__(self, image, labels=None, return_inverter=False):
        self.resize.interpolation_mode = np.random.choice(self.interpolation_modes)
        self.resize.labels_format = self.labels_format
        return self.resize(image, labels, return_inverter)

class Flip:
    '''
    Flips images horizontally or vertically.
    '''
    def __init__(self,
                 dim='horizontal',
                 labels_format={'class_id': 0, 'xmin': 1, 'ymin': 2, 'xmax': 3, 'ymax': 4}):
        '''
        Arguments:
            dim (str, optional): Can be either of 'horizontal' and 'vertical'.
                If 'horizontal', images will be flipped horizontally, i.e. along
                the vertical axis. If 'horizontal', images will be flipped vertically,
                i.e. along the horizontal axis.
            labels_format (dict, optional): A dictionary that defines which index in the last axis of the labels
                of an image contains which bounding box coordinate. The dictionary maps at least the keywords
                'xmin', 'ymin', 'xmax', and 'ymax' to their respective indices within last axis of the labels array.
        '''
        if not (dim in {'horizontal', 'vertical'}): raise ValueError("`dim` can be one of 'horizontal' and 'vertical'.")
        self.dim = dim
        self.labels_format = labels_format

    def __call__(self, image, labels=None, return_inverter=False):

        img_height, img_width = image.shape[:2]

        xmin = self.labels_format['xmin']
        ymin = self.labels_format['ymin']
        xmax = self.labels_format['xmax']
        ymax = self.labels_format['ymax']

        if self.dim == 'horizontal':
            image = image[:,::-1]
            if labels is None:
                return image
            else:
                labels = np.copy(labels)
                labels[:, [xmin, xmax]] = img_width - labels[:, [xmax, xmin]]
                return image, labels
        else:
            image = image[::-1]
            if labels is None:
                return image
            else:
                labels = np.copy(labels)
                labels[:, [ymin, ymax]] = img_height - labels[:, [ymax, ymin]]
                return image, labels

class RandomFlip:
    '''
    Randomly flips images horizontally or vertically. The randomness only refers
    to whether or not the image will be flipped.
    '''
    def __init__(self,
                 dim='horizontal',
                 prob=0.5,
                 labels_format={'class_id': 0, 'xmin': 1, 'ymin': 2, 'xmax': 3, 'ymax': 4}):
        '''
        Arguments:
            dim (str, optional): Can be either of 'horizontal' and 'vertical'.
                If 'horizontal', images will be flipped horizontally, i.e. along
                the vertical axis. If 'horizontal', images will be flipped vertically,
                i.e. along the horizontal axis.
            prob (float, optional): `(1 - prob)` determines the probability with which the original,
                unaltered image is returned.
            labels_format (dict, optional): A dictionary that defines which index in the last axis of the labels
                of an image contains which bounding box coordinate. The dictionary maps at least the keywords
                'xmin', 'ymin', 'xmax', and 'ymax' to their respective indices within last axis of the labels array.
        '''
        self.dim = dim
        self.prob = prob
        self.labels_format = labels_format
        self.flip = Flip(dim=self.dim, labels_format=self.labels_format)

    def __call__(self, image, labels=None):
        p = np.random.uniform(0,1)
        if p >= (1.0-self.prob):
            self.flip.labels_format = self.labels_format
            return self.flip(image, labels)
        elif labels is None:
            return image
        else:
            return image, labels

class Translate:
    '''
    Translates images horizontally and/or vertically.
    '''

    def __init__(self,
                 dy,
                 dx,
                 clip_boxes=True,
                 box_filter=None,
                 background=(0,0,0),
                 labels_format={'class_id': 0, 'xmin': 1, 'ymin': 2, 'xmax': 3, 'ymax': 4}):
        '''
        Arguments:
            dy (float): The fraction of the image height by which to translate images along the
                vertical axis. Positive values translate images downwards, negative values
                translate images upwards.
            dx (float): The fraction of the image width by which to translate images along the
                horizontal axis. Positive values translate images to the right, negative values
                translate images to the left.
            clip_boxes (bool, optional): Only relevant if ground truth bounding boxes are given.
                If `True`, any ground truth bounding boxes will be clipped to lie entirely within the
                image after the translation.
            box_filter (BoxFilter, optional): Only relevant if ground truth bounding boxes are given.
                A `BoxFilter` object to filter out bounding boxes that don't meet the given criteria
                after the transformation. Refer to the `BoxFilter` documentation for details. If `None`,
                the validity of the bounding boxes is not checked.
            background (list/tuple, optional): A 3-tuple specifying the RGB color value of the
                background pixels of the translated images.
            labels_format (dict, optional): A dictionary that defines which index in the last axis of the labels
                of an image contains which bounding box coordinate. The dictionary maps at least the keywords
                'xmin', 'ymin', 'xmax', and 'ymax' to their respective indices within last axis of the labels array.
        '''

        if not (isinstance(box_filter, BoxFilter) or box_filter is None):
            raise ValueError("`box_filter` must be either `None` or a `BoxFilter` object.")
        self.dy_rel = dy
        self.dx_rel = dx
        self.clip_boxes = clip_boxes
        self.box_filter = box_filter
        self.background = background
        self.labels_format = labels_format

    def __call__(self, image, labels=None):

        img_height, img_width = image.shape[:2]

        # Compute the translation matrix.
        dy_abs = int(round(img_height * self.dy_rel))
        dx_abs = int(round(img_width * self.dx_rel))
        M = np.float32([[1, 0, dx_abs],
                        [0, 1, dy_abs]])

        # Translate the image.
        image = cv2.warpAffine(image,
                               M=M,
                               dsize=(img_width, img_height),
                               borderMode=cv2.BORDER_CONSTANT,
                               borderValue=self.background)

        if labels is None:
            return image
        else:
            xmin = self.labels_format['xmin']
            ymin = self.labels_format['ymin']
            xmax = self.labels_format['xmax']
            ymax = self.labels_format['ymax']

            labels = np.copy(labels)
            # Translate the box coordinates to the translated image's coordinate system.
            labels[:,[xmin,xmax]] += dx_abs
            labels[:,[ymin,ymax]] += dy_abs

            # Compute all valid boxes for this patch.
            if not (self.box_filter is None):
                self.box_filter.labels_format = self.labels_format
                labels = self.box_filter(labels=labels,
                                         image_height=img_height,
                                         image_width=img_width)

            if self.clip_boxes:
                labels[:,[ymin,ymax]] = np.clip(labels[:,[ymin,ymax]], a_min=0, a_max=img_height-1)
                labels[:,[xmin,xmax]] = np.clip(labels[:,[xmin,xmax]], a_min=0, a_max=img_width-1)

            return image, labels

class RandomTranslate:
    '''
    Randomly translates images horizontally and/or vertically.
    '''

    def __init__(self,
                 dy_minmax=(0.03,0.3),
                 dx_minmax=(0.03,0.3),
                 prob=0.5,
                 clip_boxes=True,
                 box_filter=None,
                 image_validator=None,
                 n_trials_max=3,
                 background=(0,0,0),
                 labels_format={'class_id': 0, 'xmin': 1, 'ymin': 2, 'xmax': 3, 'ymax': 4}):
        '''
        Arguments:
            dy_minmax (list/tuple, optional): A 2-tuple `(min, max)` of non-negative floats that
                determines the minimum and maximum relative translation of images along the vertical
                axis both upward and downward. That is, images will be randomly translated by at least
                `min` and at most `max` either upward or downward. For example, if `dy_minmax == (0.05,0.3)`,
                an image of size `(100,100)` will be translated by at least 5 and at most 30 pixels
                either upward or downward. The translation direction is chosen randomly.
            dx_minmax (list/tuple, optional): A 2-tuple `(min, max)` of non-negative floats that
                determines the minimum and maximum relative translation of images along the horizontal
                axis both to the left and right. That is, images will be randomly translated by at least
                `min` and at most `max` either left or right. For example, if `dx_minmax == (0.05,0.3)`,
                an image of size `(100,100)` will be translated by at least 5 and at most 30 pixels
                either left or right. The translation direction is chosen randomly.
            prob (float, optional): `(1 - prob)` determines the probability with which the original,
                unaltered image is returned.
            clip_boxes (bool, optional): Only relevant if ground truth bounding boxes are given.
                If `True`, any ground truth bounding boxes will be clipped to lie entirely within the
                image after the translation.
            box_filter (BoxFilter, optional): Only relevant if ground truth bounding boxes are given.
                A `BoxFilter` object to filter out bounding boxes that don't meet the given criteria
                after the transformation. Refer to the `BoxFilter` documentation for details. If `None`,
                the validity of the bounding boxes is not checked.
            image_validator (ImageValidator, optional): Only relevant if ground truth bounding boxes are given.
                An `ImageValidator` object to determine whether a translated image is valid. If `None`,
                any outcome is valid.
            n_trials_max (int, optional): Only relevant if ground truth bounding boxes are given.
                Determines the maxmial number of trials to produce a valid image. If no valid image could
                be produced in `n_trials_max` trials, returns the unaltered input image.
            background (list/tuple, optional): A 3-tuple specifying the RGB color value of the
                background pixels of the translated images.
            labels_format (dict, optional): A dictionary that defines which index in the last axis of the labels
                of an image contains which bounding box coordinate. The dictionary maps at least the keywords
                'xmin', 'ymin', 'xmax', and 'ymax' to their respective indices within last axis of the labels array.
        '''
        if dy_minmax[0] > dy_minmax[1]:
            raise ValueError("It must be `dy_minmax[0] <= dy_minmax[1]`.")
        if dx_minmax[0] > dx_minmax[1]:
            raise ValueError("It must be `dx_minmax[0] <= dx_minmax[1]`.")
        if dy_minmax[0] < 0 or dx_minmax[0] < 0:
            raise ValueError("It must be `dy_minmax[0] >= 0` and `dx_minmax[0] >= 0`.")
        if not (isinstance(image_validator, ImageValidator) or image_validator is None):
            raise ValueError("`image_validator` must be either `None` or an `ImageValidator` object.")
        self.dy_minmax = dy_minmax
        self.dx_minmax = dx_minmax
        self.prob = prob
        self.clip_boxes = clip_boxes
        self.box_filter = box_filter
        self.image_validator = image_validator
        self.n_trials_max = n_trials_max
        self.background = background
        self.labels_format = labels_format
        self.translate = Translate(dy=0,
                                   dx=0,
                                   clip_boxes=self.clip_boxes,
                                   box_filter=self.box_filter,
                                   background=self.background,
                                   labels_format=self.labels_format)

    def __call__(self, image, labels=None):

        p = np.random.uniform(0,1)
        if p >= (1.0-self.prob):

            img_height, img_width = image.shape[:2]

            xmin = self.labels_format['xmin']
            ymin = self.labels_format['ymin']
            xmax = self.labels_format['xmax']
            ymax = self.labels_format['ymax']

            # Override the preset labels format.
            if not self.image_validator is None:
                self.image_validator.labels_format = self.labels_format
            self.translate.labels_format = self.labels_format

            for _ in range(max(1, self.n_trials_max)):

                # Pick the relative amount by which to translate.
                dy_abs = np.random.uniform(self.dy_minmax[0], self.dy_minmax[1])
                dx_abs = np.random.uniform(self.dx_minmax[0], self.dx_minmax[1])
                # Pick the direction in which to translate.
                dy = np.random.choice([-dy_abs, dy_abs])
                dx = np.random.choice([-dx_abs, dx_abs])
                self.translate.dy_rel = dy
                self.translate.dx_rel = dx

                if (labels is None) or (self.image_validator is None):
                    # We either don't have any boxes or if we do, we will accept any outcome as valid.
                    return self.translate(image, labels)
                else:
                    # Translate the box coordinates to the translated image's coordinate system.
                    new_labels = np.copy(labels)
                    new_labels[:, [ymin, ymax]] += int(round(img_height * dy))
                    new_labels[:, [xmin, xmax]] += int(round(img_width * dx))

                    # Check if the patch is valid.
                    if self.image_validator(labels=new_labels,
                                            image_height=img_height,
                                            image_width=img_width):
                        return self.translate(image, labels)

            # If all attempts failed, return the unaltered input image.
            if labels is None:
                return image

            else:
                return image, labels

        elif labels is None:
            return image

        else:
            return image, labels

class Scale:
    '''
    Scales images, i.e. zooms in or out.
    '''

    def __init__(self,
                 factor,
                 clip_boxes=True,
                 box_filter=None,
                 background=(0,0,0),
                 labels_format={'class_id': 0, 'xmin': 1, 'ymin': 2, 'xmax': 3, 'ymax': 4}):
        '''
        Arguments:
            factor (float): The fraction of the image size by which to scale images. Must be positive.
            clip_boxes (bool, optional): Only relevant if ground truth bounding boxes are given.
                If `True`, any ground truth bounding boxes will be clipped to lie entirely within the
                image after the translation.
            box_filter (BoxFilter, optional): Only relevant if ground truth bounding boxes are given.
                A `BoxFilter` object to filter out bounding boxes that don't meet the given criteria
                after the transformation. Refer to the `BoxFilter` documentation for details. If `None`,
                the validity of the bounding boxes is not checked.
            background (list/tuple, optional): A 3-tuple specifying the RGB color value of the potential
                background pixels of the scaled images.
            labels_format (dict, optional): A dictionary that defines which index in the last axis of the labels
                of an image contains which bounding box coordinate. The dictionary maps at least the keywords
                'xmin', 'ymin', 'xmax', and 'ymax' to their respective indices within last axis of the labels array.
        '''

        if factor <= 0:
            raise ValueError("It must be `factor > 0`.")
        if not (isinstance(box_filter, BoxFilter) or box_filter is None):
            raise ValueError("`box_filter` must be either `None` or a `BoxFilter` object.")
        self.factor = factor
        self.clip_boxes = clip_boxes
        self.box_filter = box_filter
        self.background = background
        self.labels_format = labels_format

    def __call__(self, image, labels=None):

        img_height, img_width = image.shape[:2]

        # Compute the rotation matrix.
        M = cv2.getRotationMatrix2D(center=(img_width / 2, img_height / 2),
                                    angle=0,
                                    scale=self.factor)

        # Scale the image.
        image = cv2.warpAffine(image,
                               M=M,
                               dsize=(img_width, img_height),
                               borderMode=cv2.BORDER_CONSTANT,
                               borderValue=self.background)

        if labels is None:
            return image
        else:
            xmin = self.labels_format['xmin']
            ymin = self.labels_format['ymin']
            xmax = self.labels_format['xmax']
            ymax = self.labels_format['ymax']

            labels = np.copy(labels)
            # Scale the bounding boxes accordingly.
            # Transform two opposite corner points of the rectangular boxes using the rotation matrix `M`.
            toplefts = np.array([labels[:,xmin], labels[:,ymin], np.ones(labels.shape[0])])
            bottomrights = np.array([labels[:,xmax], labels[:,ymax], np.ones(labels.shape[0])])
            new_toplefts = (np.dot(M, toplefts)).T
            new_bottomrights = (np.dot(M, bottomrights)).T
            labels[:,[xmin,ymin]] = np.round(new_toplefts, decimals=0).astype(np.int)
            labels[:,[xmax,ymax]] = np.round(new_bottomrights, decimals=0).astype(np.int)

            # Compute all valid boxes for this patch.
            if not (self.box_filter is None):
                self.box_filter.labels_format = self.labels_format
                labels = self.box_filter(labels=labels,
                                         image_height=img_height,
                                         image_width=img_width)

            if self.clip_boxes:
                labels[:,[ymin,ymax]] = np.clip(labels[:,[ymin,ymax]], a_min=0, a_max=img_height-1)
                labels[:,[xmin,xmax]] = np.clip(labels[:,[xmin,xmax]], a_min=0, a_max=img_width-1)

            return image, labels

class RandomScale:
    '''
    Randomly scales images.
    '''

    def __init__(self,
                 min_factor=0.5,
                 max_factor=1.5,
                 prob=0.5,
                 clip_boxes=True,
                 box_filter=None,
                 image_validator=None,
                 n_trials_max=3,
                 background=(0,0,0),
                 labels_format={'class_id': 0, 'xmin': 1, 'ymin': 2, 'xmax': 3, 'ymax': 4}):
        '''
        Arguments:
            min_factor (float, optional): The minimum fraction of the image size by which to scale images.
                Must be positive.
            max_factor (float, optional): The maximum fraction of the image size by which to scale images.
                Must be positive.
            prob (float, optional): `(1 - prob)` determines the probability with which the original,
                unaltered image is returned.
            clip_boxes (bool, optional): Only relevant if ground truth bounding boxes are given.
                If `True`, any ground truth bounding boxes will be clipped to lie entirely within the
                image after the translation.
            box_filter (BoxFilter, optional): Only relevant if ground truth bounding boxes are given.
                A `BoxFilter` object to filter out bounding boxes that don't meet the given criteria
                after the transformation. Refer to the `BoxFilter` documentation for details. If `None`,
                the validity of the bounding boxes is not checked.
            image_validator (ImageValidator, optional): Only relevant if ground truth bounding boxes are given.
                An `ImageValidator` object to determine whether a scaled image is valid. If `None`,
                any outcome is valid.
            n_trials_max (int, optional): Only relevant if ground truth bounding boxes are given.
                Determines the maxmial number of trials to produce a valid image. If no valid image could
                be produced in `n_trials_max` trials, returns the unaltered input image.
            background (list/tuple, optional): A 3-tuple specifying the RGB color value of the potential
                background pixels of the scaled images.
            labels_format (dict, optional): A dictionary that defines which index in the last axis of the labels
                of an image contains which bounding box coordinate. The dictionary maps at least the keywords
                'xmin', 'ymin', 'xmax', and 'ymax' to their respective indices within last axis of the labels array.
        '''

        if not (0 < min_factor <= max_factor):
            raise ValueError("It must be `0 < min_factor <= max_factor`.")
        if not (isinstance(image_validator, ImageValidator) or image_validator is None):
            raise ValueError("`image_validator` must be either `None` or an `ImageValidator` object.")
        self.min_factor = min_factor
        self.max_factor = max_factor
        self.prob = prob
        self.clip_boxes = clip_boxes
        self.box_filter = box_filter
        self.image_validator = image_validator
        self.n_trials_max = n_trials_max
        self.background = background
        self.labels_format = labels_format
        self.scale = Scale(factor=1.0,
                           clip_boxes=self.clip_boxes,
                           box_filter=self.box_filter,
                           background=self.background,
                           labels_format=self.labels_format)

    def __call__(self, image, labels=None):

        p = np.random.uniform(0,1)
        if p >= (1.0-self.prob):

            img_height, img_width = image.shape[:2]

            xmin = self.labels_format['xmin']
            ymin = self.labels_format['ymin']
            xmax = self.labels_format['xmax']
            ymax = self.labels_format['ymax']

            # Override the preset labels format.
            if not self.image_validator is None:
                self.image_validator.labels_format = self.labels_format
            self.scale.labels_format = self.labels_format

            for _ in range(max(1, self.n_trials_max)):

                # Pick a scaling factor.
                factor = np.random.uniform(self.min_factor, self.max_factor)
                self.scale.factor = factor

                if (labels is None) or (self.image_validator is None):
                    # We either don't have any boxes or if we do, we will accept any outcome as valid.
                    return self.scale(image, labels)
                else:
                    # Scale the bounding boxes accordingly.
                    # Transform two opposite corner points of the rectangular boxes using the rotation matrix `M`.
                    toplefts = np.array([labels[:,xmin], labels[:,ymin], np.ones(labels.shape[0])])
                    bottomrights = np.array([labels[:,xmax], labels[:,ymax], np.ones(labels.shape[0])])

                    # Compute the rotation matrix.
                    M = cv2.getRotationMatrix2D(center=(img_width / 2, img_height / 2),
                                                angle=0,
                                                scale=factor)

                    new_toplefts = (np.dot(M, toplefts)).T
                    new_bottomrights = (np.dot(M, bottomrights)).T

                    new_labels = np.copy(labels)
                    new_labels[:,[xmin,ymin]] = np.around(new_toplefts, decimals=0).astype(np.int)
                    new_labels[:,[xmax,ymax]] = np.around(new_bottomrights, decimals=0).astype(np.int)

                    # Check if the patch is valid.
                    if self.image_validator(labels=new_labels,
                                            image_height=img_height,
                                            image_width=img_width):
                        return self.scale(image, labels)

            # If all attempts failed, return the unaltered input image.
            if labels is None:
                return image

            else:
                return image, labels

        elif labels is None:
            return image

        else:
            return image, labels

class Rotate:
    '''
    Rotates images counter-clockwise by 90, 180, or 270 degrees.
    '''

    def __init__(self,
                 angle,
                 labels_format={'class_id': 0, 'xmin': 1, 'ymin': 2, 'xmax': 3, 'ymax': 4}):
        '''
        Arguments:
            angle (int): The angle in degrees by which to rotate the images counter-clockwise.
                Only 90, 180, and 270 are valid values.
            labels_format (dict, optional): A dictionary that defines which index in the last axis of the labels
                of an image contains which bounding box coordinate. The dictionary maps at least the keywords
                'xmin', 'ymin', 'xmax', and 'ymax' to their respective indices within last axis of the labels array.
        '''

        if not angle in {90, 180, 270}:
            raise ValueError("`angle` must be in the set {90, 180, 270}.")
        self.angle = angle
        self.labels_format = labels_format

    def __call__(self, image, labels=None):

        img_height, img_width = image.shape[:2]

        # Compute the rotation matrix.
        M = cv2.getRotationMatrix2D(center=(img_width / 2, img_height / 2),
                                    angle=self.angle,
                                    scale=1)

        # Get the sine and cosine from the rotation matrix.
        cos_angle = np.abs(M[0, 0])
        sin_angle = np.abs(M[0, 1])

        # Compute the new bounding dimensions of the image.
        img_width_new = int(img_height * sin_angle + img_width * cos_angle)
        img_height_new = int(img_height * cos_angle + img_width * sin_angle)

        # Adjust the rotation matrix to take into account the translation.
        M[1, 2] += (img_height_new - img_height) / 2
        M[0, 2] += (img_width_new - img_width) / 2

        # Rotate the image.
        image = cv2.warpAffine(image,
                               M=M,
                               dsize=(img_width_new, img_height_new))

        if labels is None:
            return image
        else:
            xmin = self.labels_format['xmin']
            ymin = self.labels_format['ymin']
            xmax = self.labels_format['xmax']
            ymax = self.labels_format['ymax']

            labels = np.copy(labels)
            # Rotate the bounding boxes accordingly.
            # Transform two opposite corner points of the rectangular boxes using the rotation matrix `M`.
            toplefts = np.array([labels[:,xmin], labels[:,ymin], np.ones(labels.shape[0])])
            bottomrights = np.array([labels[:,xmax], labels[:,ymax], np.ones(labels.shape[0])])
            new_toplefts = (np.dot(M, toplefts)).T
            new_bottomrights = (np.dot(M, bottomrights)).T
            labels[:,[xmin,ymin]] = np.round(new_toplefts, decimals=0).astype(np.int)
            labels[:,[xmax,ymax]] = np.round(new_bottomrights, decimals=0).astype(np.int)

            if self.angle == 90:
                # ymin and ymax were switched by the rotation.
                labels[:,[ymax,ymin]] = labels[:,[ymin,ymax]]
            elif self.angle == 180:
                # ymin and ymax were switched by the rotation,
                # and also xmin and xmax were switched.
                labels[:,[ymax,ymin]] = labels[:,[ymin,ymax]]
                labels[:,[xmax,xmin]] = labels[:,[xmin,xmax]]
            elif self.angle == 270:
                # xmin and xmax were switched by the rotation.
                labels[:,[xmax,xmin]] = labels[:,[xmin,xmax]]

            return image, labels

class RandomRotate:
    '''
    Randomly rotates images counter-clockwise.
    '''

    def __init__(self,
                 angles=[90, 180, 270],
                 prob=0.5,
                 labels_format={'class_id': 0, 'xmin': 1, 'ymin': 2, 'xmax': 3, 'ymax': 4}):
        '''
        Arguments:
            angle (list): The list of angles in degrees from which one is randomly selected to rotate
                the images counter-clockwise. Only 90, 180, and 270 are valid values.
            prob (float, optional): `(1 - prob)` determines the probability with which the original,
                unaltered image is returned.
            labels_format (dict, optional): A dictionary that defines which index in the last axis of the labels
                of an image contains which bounding box coordinate. The dictionary maps at least the keywords
                'xmin', 'ymin', 'xmax', and 'ymax' to their respective indices within last axis of the labels array.
        '''
        for angle in angles:
            if not angle in {90, 180, 270}:
                raise ValueError("`angles` can only contain the values 90, 180, and 270.")
        self.angles = angles
        self.prob = prob
        self.labels_format = labels_format
        self.rotate = Rotate(angle=90, labels_format=self.labels_format)

    def __call__(self, image, labels=None):

        p = np.random.uniform(0,1)
        if p >= (1.0-self.prob):
            # Pick a rotation angle.
            self.rotate.angle = random.choice(self.angles)
            self.rotate.labels_format = self.labels_format
            return self.rotate(image, labels)

        elif labels is None:
            return image

        else:
            return image, labels



"""Utilities for HRA data preprocessing, prediction decoding and plotting.
"""

from __future__ import print_function
import numpy as np
import matplotlib.pyplot as plt

from keras.preprocessing import image
from keras.applications.inception_v3 import preprocess_input

from keras.utils import get_file
import json
from utils.generic_utils import imagenet_preprocess_input, places_preprocess_input


target_size = (224, 224)

CLASS_INDEX = None
CL_CLASS_INDEX_PATH = 'https://github.com/GKalliatakis/ubiquitous-assets/releases/download/v0.1.3/HRA_2classCL_index.json'
DP_CLASS_INDEX_PATH = 'https://github.com/GKalliatakis/ubiquitous-assets/releases/download/v0.1.3/HRA_2classDP_index.json'


def _obtain_train_mode(nb_of_conv_layers_to_fine_tune):
    """Obtains the train mode string based on the provided number of conv. layers that will be fine-tuned.

    # Arguments
        nb_of_conv_layers_to_fine_tune: integer to indicate the number of convolutional layersto fine-tune.
            One of `1`, `2` or `3`.
    # Returns
        A string that will serve as the train mode of the model.
    """

    if nb_of_conv_layers_to_fine_tune == None:
        return 'feature_extraction'
    elif nb_of_conv_layers_to_fine_tune in {1, 2, 3}:
        return 'fine_tuning'
    else:
        raise ValueError('The `nb_of_conv_layers_to_fine_tune` argument should be either '
                         '`None` (indicates feature extraction mode), '
                         '`1`, `2` or `3`. More than 3 conv. blocks are not included '
                         'because the more parameters we are training (unfreezing), the more we are at risk of overfitting.')



def _obtain_first_phase_trained_weights (violation_class,
                                         model_name):
    """Retrieves the weights of an already trained feature extraction model.
        Only relevant when using `fine_tuning` as train_mode after `feature_extraction` weights have been saved.

    # Arguments
        violation_class: violation_class: one of `cl` (HRA dataset with 2 classes - [i]'child_labour' and [ii]'no violation')
            or `dp` (HRA dataset with 2 classes - [i]'displaced_populations' and [ii]'no violation').
        model_name: String to declare the name of the model
    # Returns
        A string with the weights path.
    """

    if violation_class == 'cl':
        first_phase_trained_weights_filename = 'trained_models/' + 'cl_' + model_name + '_weights_feature_extraction_tf_dim_ordering_tf_kernels.h5'
    elif violation_class == 'dp':
        first_phase_trained_weights_filename = 'trained_models/' + 'dp_' + model_name + '_weights_feature_extraction_tf_dim_ordering_tf_kernels.h5'


    return first_phase_trained_weights_filename





def _obtain_weights_CSVLogger_filenames(violation_class,
                                        train_mode,
                                        model_name,
                                        nb_of_conv_layers_to_fine_tune):
    """Obtains the polished filenames for the weights and the CSVLogger of the model.

    # Arguments
        violation_class: violation_class: one of `cl` (HRA dataset with 2 classes - [i]'child_labour' and [ii]'no violation')
            or `dp` (HRA dataset with 2 classes - [i]'displaced_populations' and [ii]'no violation').
        train_mode: String to declare the train mode of the model (how many layers will be frozen during training).
            - `feature_extraction` taking the convolutional base of a previously-trained network,
                running the new data through it, and training a new classifier on top of the output.
            - `fine_tuning` unfreezing a few of the top layers of a frozen conv. base used for feature extraction,
                and jointly training both the newly added part of the model and these top layers.
        model_name: String to declare the name of the model
        nb_of_conv_layers_to_fine_tune: integer to indicate the number of convolutional
            layers to fine-tune. One of `1`, `2` or `3`.

    # Returns
        Two strings that will serve as the filenames for the weights and the CSVLogger respectively.
    """

    if violation_class == 'cl':
        if train_mode == 'feature_extraction':

            weights_filename = 'trained_models/' + 'cl_' + model_name + '_weights_feature_extraction_tf_dim_ordering_tf_kernels.h5'
            CSVLogger_filename = model_name + '_cl_feature_extraction.csv'

        else:

            if nb_of_conv_layers_to_fine_tune == 1:
                weights_filename = 'trained_models/' + 'cl_' + model_name + '_weights_one_layer_tf_dim_ordering_tf_kernels.h5'
                CSVLogger_filename = model_name + '_cl_fine_tuning_one_layer.csv'

            elif nb_of_conv_layers_to_fine_tune == 2:
                weights_filename = 'trained_models/' + 'cl_' + model_name + '_weights_two_layers_tf_dim_ordering_tf_kernels.h5'
                CSVLogger_filename = model_name + '_cl_fine_tuning_two_layers.csv'

            elif nb_of_conv_layers_to_fine_tune == 3:
                weights_filename = 'trained_models/' + 'cl_' + model_name + '_weights_three_layers_tf_dim_ordering_tf_kernels.h5'
                CSVLogger_filename = model_name + '_cl_fine_tuning_three_layers.csv'

            else:
                raise NotImplementedError(
                    'The `nb_of_conv_layers_to_fine_tune` argument should be either `1`, `2` or `3`. '
                    'More than 3 conv. blocks are not supported because the more parameters we are training, '
                    'the more we are at risk of overfitting.')

    elif violation_class == 'dp':
        if train_mode == 'feature_extraction':

            weights_filename = 'trained_models/' + 'dp_' + model_name + '_weights_feature_extraction_tf_dim_ordering_tf_kernels.h5'
            CSVLogger_filename = model_name + '_dp_feature_extraction.csv'

        else:

            if nb_of_conv_layers_to_fine_tune == 1:
                weights_filename = 'trained_models/' + 'dp_' + model_name + '_weights_one_layer_tf_dim_ordering_tf_kernels.h5'
                CSVLogger_filename = model_name + '_dp_fine_tuning_one_layer.csv'

            elif nb_of_conv_layers_to_fine_tune == 2:
                weights_filename = 'trained_models/' + 'dp_' + model_name + '_weights_two_layers_tf_dim_ordering_tf_kernels.h5'
                CSVLogger_filename = model_name + '_dp_fine_tuning_two_layers.csv'

            elif nb_of_conv_layers_to_fine_tune == 3:
                weights_filename = 'trained_models/' + 'dp_' + model_name + '_weights_three_layers_tf_dim_ordering_tf_kernels.h5'
                CSVLogger_filename = model_name + '_dp_cl_fine_tuning_three_layers.csv'

            else:
                raise NotImplementedError(
                    'The `nb_of_conv_layers_to_fine_tune` argument should be either `1`, `2` or `3`. '
                    'More than 3 conv. blocks are not supported because the more parameters we are training, '
                    'the more we are at risk of overfitting.')


    return weights_filename, CSVLogger_filename




def _obtain_weights_path(violation_class,
                         pre_trained_model,
                         nb_of_conv_layers_to_fine_tune,
                         include_top):
    """Obtains the polished filenames for the weights of a trained Keras model.

        # Arguments
            violation_class: violation_class: one of `cl` (HRA dataset with 2 classes - [i]'child_labour' and [ii]'no violation')
                or `dp` (HRA dataset with 2 classes - [i]'displaced_populations' and [ii]'no violation').
            pre_trained_model: 'One of `vgg16`, `vgg19`, `resnet50` or `vgg16_places365`.
            nb_of_conv_layers_to_fine_tune: integer to indicate the number of convolutional
                layers to fine-tune. One of `1`, `2` or `3`.
            include_top: whether to include the 3 fully-connected layers at the top of the network.

        # Returns
            Two strings that will serve as the original URL of the file (origin) and name of the file (fname) for loading the weights.
        """

    # This is the only URL that must be altered in case of changing the repo where weights files are stored.
    # The rest will be automatically inferred from the github_repo URL.
    # Note that the structure of the releases must comply with the following structure:
    # main_release_dir/
    #               v0.1.1(weights for feature extraction mode)/
    #               v0.1.2(weights for fine-tuning mode)/


    github_repo = 'https://github.com/GKalliatakis/ubiquitous-assets/releases'

    if nb_of_conv_layers_to_fine_tune == None:
        fname = violation_class + '_' + pre_trained_model + '_weights_feature_extraction_tf_dim_ordering_tf_kernels.h5'

    elif nb_of_conv_layers_to_fine_tune == 1:
        if include_top:
            fname = violation_class + '_' + pre_trained_model + '_weights_one_layer_tf_dim_ordering_tf_kernels.h5'
        else:
            fname = violation_class + '_' + pre_trained_model + '_weights_one_layer_tf_dim_ordering_tf_kernels_notop.h5'

    elif nb_of_conv_layers_to_fine_tune == 2:
        if include_top:
            fname = violation_class + '_' + pre_trained_model + '_weights_two_layers_tf_dim_ordering_tf_kernels.h5'
        else:
            fname = violation_class + '_' + pre_trained_model + '_weights_two_layers_tf_dim_ordering_tf_kernels_notop.h5'

    elif nb_of_conv_layers_to_fine_tune == 3:
        if include_top:
            fname = violation_class + '_' + pre_trained_model + '_weights_three_layers_tf_dim_ordering_tf_kernels.h5'
        else:
            fname = violation_class + '_' + pre_trained_model + '_weights_three_layers_tf_dim_ordering_tf_kernels_notop.h5'



    if nb_of_conv_layers_to_fine_tune == None:
        origin = github_repo + '/download/v0.1.1/' + fname
    else:
        origin = github_repo + '/download/v0.1.2/' + fname


    return origin, fname



def decode_predictions(violation_class,preds, top=2):
    """Decodes the prediction of a HRA-2CLASS model.

        # Arguments
            violation_class: violation_class: one of `cl` (HRA dataset with 2 classes - [i]'child_labour' and [ii]'no violation')
                or `dp` (HRA dataset with 2 classes - [i]'displaced_populations' and [ii]'no violation')
            preds: Numpy tensor encoding a batch of predictions.
            top: integer, how many top-guesses to return.

        # Returns
            A list of lists of top class prediction tuples `(class_name, class_description, score)`.
            One list of tuples per sample in batch input.

        # Raises
            ValueError: in case of invalid shape of the `pred` array must be 2D).
    """
    global CLASS_INDEX
    if len(preds.shape) != 2 or preds.shape[1] != 2:
        raise ValueError('`decode_predictions` expects '
                         'a batch of predictions '
                         '(i.e. a 2D array of shape (samples, 2)). '
                         'Found array with shape: ' + str(preds.shape))
    if CLASS_INDEX is None:
        if violation_class =='cl':
            fpath = get_file('HRA_2classCL_index.json',
                             CL_CLASS_INDEX_PATH,
                             cache_subdir='AbuseNet')
            CLASS_INDEX = json.load(open(fpath))
        elif violation_class =='dp':
            fpath = get_file('HRA_2classDP_index.json',
                             DP_CLASS_INDEX_PATH,
                             cache_subdir='AbuseNet')
            CLASS_INDEX = json.load(open(fpath))

    results = []
    for pred in preds:
        top_indices = pred.argsort()[-top:][::-1]
        result = [tuple(CLASS_INDEX[str(i)]) + (pred[i],) for i in top_indices]
        result.sort(key=lambda x: x[2], reverse=True)
        results.append(result)

    return results


def predict(violation_class, model, img, target_size):
  """Generates output predictions for a single PIL image.

    # Arguments
        violation_class: violation_class: one of `cl` (HRA dataset with 2 classes - [i]'child_labour' and [ii]'no violation')
            or `dp` (HRA dataset with 2 classes - [i]'displaced_populations' and [ii]'no violation')
        model: keras model
        img: PIL format image
        target_size: (w,h) tuple

    # Returns
        list of predicted labels and their probabilities
  """
  if img.size != target_size:
    img = img.resize(target_size)

  x = image.img_to_array(img)
  x = np.expand_dims(x, axis=0)
  x = preprocess_input(x)
  preds = model.predict(x)

  # print ('Raw preds: ',preds )

  return preds, decode_predictions(violation_class = violation_class, preds = preds, top=2)[0]


def predict_v2(violation_class, model, img, target_size):
  """Generates output predictions for a single PIL image.
    # Arguments
        violation_class: violation_class: one of `cl` (HRA dataset with 2 classes - [i]'child_labour' and [ii]'no violation')
            or `dp` (HRA dataset with 2 classes - [i]'displaced_populations' and [ii]'no violation')
        model: keras model
        img: PIL format image
        target_size: (w,h) tuple
    # Returns
        list of predicted labels and their probabilities
  """
  if img.size != target_size:
    img = img.resize(target_size)

  x = image.img_to_array(img)
  x = np.expand_dims(x, axis=0)
  x = preprocess_input(x)
  preds = model.predict(x)

  return decode_predictions(violation_class = violation_class, preds = preds, top=2)[0]


def duo_ensemble_predict(violation_class,
                         model_a, model_b,
                         img,
                         target_size
                         ):
  """Generates output predictions for a single PIL image for 2 different models,
    and then puts together those predictions by averaging them at inference time.

    # Arguments
        violation_class: violation_class: one of `cl` (HRA dataset with 2 classes - [i]'child_labour' and [ii]'no violation')
            or `dp` (HRA dataset with 2 classes - [i]'displaced_populations' and [ii]'no violation')
        model_a: 1st model
        model_b: 2nd model
        img: PIL format image
        target_size: (w,h) tuple

    # Returns
        list of predicted labels (which have been pooled accordingly) and their probabilities
  """
  if img.size != target_size:
    img = img.resize(target_size)

  x = image.img_to_array(img)
  x = np.expand_dims(x, axis=0)
  x = preprocess_input(x)

  preds_a = model_a.predict(x)
  preds_b = model_b.predict(x)
  final_preds = 0.50 * (preds_a + preds_b)

  return decode_predictions(violation_class = violation_class, preds = final_preds, top=2)[0]



def plot_preds(violation_class, image, preds):
  """Displays image and the top-n predicted probabilities in a bar graph.

  # Arguments
    violation_class: violation_class: one of `cl` (HRA dataset with 2 classes - [i]'child_labour' and [ii]'no violation')
        or `dp` (HRA dataset with 2 classes - [i]'displaced_populations' and [ii]'no violation')
    image: PIL image
    preds: list of predicted labels and their probabilities
  """

  if violation_class == 'cl':
      labels = ("Child Labour", "NO Child Labour")

  elif violation_class == 'dp':
      labels = ("Displaced Populations", "NO Displaced Populations")



  order = list(reversed(range(len(preds))))
  plt.imshow(image)
  plt.axis('off')

  # fig = plt.figure(figsize=(2, 2))
  #
  # fig.add_subplot(1, 1, 1)
  # plt.imshow(image)
  #
  # fig.add_subplot(2, 2, 2)
  # plt.barh(order, preds, alpha=0.55)


  plt.figure()
  plt.barh(order, preds, alpha=0.55)
  plt.yticks(order, labels)
  plt.xlabel('Probability')
  plt.xlim(0,1.01)
  plt.tight_layout()
  plt.show()


def prepare_input_data(img_path,
                       objects_or_places_flag):
    """Prepares the raw images for the EMOTIC model.

    # Arguments
        body_path: Path to body only image file.
        image_path: Path to entire image file.

    # Returns
        The two processed images
    """

    body_img = image.load_img(img_path, target_size=(224, 224))
    x1 = image.img_to_array(body_img)
    x1 = np.expand_dims(x1, axis=0)

    if objects_or_places_flag == 'objects':
        x1 = imagenet_preprocess_input(x1)

    elif objects_or_places_flag == 'places':
        x1 = places_preprocess_input(x1)


    return x1


from __future__ import print_function
import os

from applications.hra_utils import prepare_input_data, predict_v2
from applications.hra_vgg16 import HRA_VGG16
from applications.hra_vgg19 import HRA_VGG19
from applications.hra_resnet50 import HRA_ResNet50
from applications.hra_vgg16_places365 import HRA_VGG16_Places365
from sklearn.metrics import accuracy_score, classification_report, precision_score, confusion_matrix, average_precision_score
from keras.preprocessing import image
from inference.displacenet_single_image_inference_unified import displaceNet_inference

class DisplaceNetBaseEvaluator(object):
    """Perfofmance metrics base class.
    """


    def __init__(self,
                 hra_model_backend_name,nb_of_conv_layers_to_fine_tune,
                 emotic_model_a_backend_name,emotic_model_b_backend_name,emotic_model_c_backend_name,
                 violation_class,
                 main_test_dir ='/home/sandbox/Desktop/Human_Rights_Archive_DB/test',
                 ):

        self.hra_model_backend_name = hra_model_backend_name
        self.nb_of_conv_layers_to_fine_tune = nb_of_conv_layers_to_fine_tune
        self.emotic_model_a_backend_name = emotic_model_a_backend_name
        self.emotic_model_b_backend_name = emotic_model_b_backend_name
        self.emotic_model_c_backend_name = emotic_model_c_backend_name
        self.main_test_dir = main_test_dir
        self.total_nb_of_test_images = sum([len(files) for r, d, files in os.walk(main_test_dir)])
        self.sorted_categories_names = sorted(os.listdir(main_test_dir))
        self.violation_class = violation_class
        self.y_true = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                       0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                       1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]



    def _obtain_y_pred(self,
                       prob_threshold=0.75):

        y_pred = []
        y_scores = []

        predicted_class_list = []
        actual_class_list = []
        coverage_count = 0

        for hra_class in self.sorted_categories_names:

            # variable that contains the main dir alongside the selected category
            tmp = os.path.join(self.main_test_dir, hra_class)
            img_names = sorted(os.listdir(tmp))

            for raw_img in img_names:
                # variable that contains the final image to be loaded
                print('Processing  [' + raw_img + ']')
                final_img = os.path.join(tmp, raw_img)

                preds = displaceNet_inference(img_path=final_img,
                                              emotic_model_a_backend_name=self.emotic_model_a_backend_name,
                                              emotic_model_b_backend_name=self.emotic_model_b_backend_name,
                                              emotic_model_c_backend_name=self.emotic_model_c_backend_name,
                                              hra_model_backend_name=self.hra_model_backend_name,
                                              nb_of_fine_tuned_conv_layers=self.nb_of_conv_layers_to_fine_tune,
                                              violation_class=self.violation_class)


                preds = preds[0]

                y_pred.append(int(preds[0][0]))
                y_scores.append(preds[0][2])

                top_1_predicted_probability = preds[0][2]

                # top_1_predicted = np.argmax(preds)
                top_1_predicted_label = preds[0][1]

                if top_1_predicted_probability >= prob_threshold:
                    coverage_count += 1

                print ('`' + hra_class + '/' + raw_img + '`  ===>  `' +
                       top_1_predicted_label + '`' + ' with ' + str(top_1_predicted_probability) + ' P')

                predicted_class_list.append(top_1_predicted_label)
                actual_class_list.append(hra_class)

        total_coverage_per = (coverage_count * 100) / self.total_nb_of_test_images

        return y_pred, self.y_true, y_scores, total_coverage_per



if __name__ == "__main__":

    violation_class = 'cl'
    hra_model_backend_name = 'VGG16'
    nb_of_conv_layers_to_fine_tune = 1

    emotic_model_a_backend_name = 'VGG19'
    emotic_model_b_backend_name = 'VGG16'
    emotic_model_c_backend_name = None

    model_backend_name = 'VGG16'

    main_test_dir = '/home/gkallia/git/AbuseNet/datasets/HRA-2clas-full-test/ChildLabour'

    # ---------------------------------------------------- #





    base_evaluator = DisplaceNetBaseEvaluator(hra_model_backend_name=hra_model_backend_name,
                                              nb_of_conv_layers_to_fine_tune=nb_of_conv_layers_to_fine_tune,
                                              emotic_model_a_backend_name=emotic_model_a_backend_name,
                                              emotic_model_b_backend_name=emotic_model_b_backend_name,
                                              emotic_model_c_backend_name=emotic_model_c_backend_name,
                                              violation_class=violation_class,
                                              main_test_dir =main_test_dir,
                                              )

    y_pred, y_true, y_scores, total_coverage_per = base_evaluator._obtain_y_pred()

    # print y_true
    top1_acc = accuracy_score(y_true, y_pred)

    AP = average_precision_score(y_true, y_scores, 'micro')


    string = model_backend_name+'-'+violation_class+'-'+str(nb_of_conv_layers_to_fine_tune)+'layer(s)'

    print('\n')
    print( '============================= %s =============================' %string)
    print(' Top-1 acc. =>  ' + str(top1_acc))
    print(' Coverage =>  ' + str(total_coverage_per) + '%')
    print(' Average Precision (AP)  =>  ' + str(AP) + '%')